@@ -124,6 +124,9 @@ func (s *snapshot) load(ctx context.Context, scopes ...interface{}) error {
 		event.Log(ctx, "go/packages.Load", tag.Snapshot.Of(s.ID()), tag.Directory.Of(cfg.Dir), tag.Query.Of(query), tag.PackageCount.Of(len(pkgs)))
 	}
 	if len(pkgs) == 0 {
+		if err == nil {
+			err = fmt.Errorf("no packages returned")
+		}
 		return errors.Errorf("%v: %w", err, source.PackagesLoadError)
 	}
 
@@ -138,7 +141,7 @@ func (s *snapshot) load(ctx context.Context, scopes ...interface{}) error {
 		}
 		// Special case for the builtin package, as it has no dependencies.
 		if pkg.PkgPath == "builtin" {
-			if err := s.view.buildBuiltinPackage(ctx, pkg.GoFiles); err != nil {
+			if err := s.buildBuiltinPackage(ctx, pkg.GoFiles); err != nil {
 				return err
 			}
 			continue
@@ -152,7 +155,7 @@ func (s *snapshot) load(ctx context.Context, scopes ...interface{}) error {
 		if err != nil {
 			return err
 		}
-		if _, err := s.buildPackageHandle(ctx, m.id, source.ParseFull); err != nil {
+		if _, err := s.buildPackageHandle(ctx, m.id, s.workspaceParseMode(m.id)); err != nil {
 			return err
 		}
 	}
