@@ -15,6 +15,7 @@ import (
 	"os"
 	"path/filepath"
 	"sort"
+	"strconv"
 	"strings"
 	"sync"
 
@@ -24,6 +25,7 @@ import (
 	"golang.org/x/tools/internal/gocommand"
 	"golang.org/x/tools/internal/lsp/debug/tag"
 	"golang.org/x/tools/internal/lsp/source"
+	"golang.org/x/tools/internal/memoize"
 	"golang.org/x/tools/internal/packagesinternal"
 	"golang.org/x/tools/internal/span"
 	"golang.org/x/tools/internal/typesinternal"
@@ -31,9 +33,17 @@ import (
 )
 
 type snapshot struct {
+	memoize.Arg // allow as a memoize.Function arg
+
 	id   uint64
 	view *View
 
+	// the cache generation that contains the data for this snapshot.
+	generation *memoize.Generation
+
+	// builtin pins the AST and package for builtin.go in memory.
+	builtin *builtinPackageHandle
+
 	// mu guards all of the maps in the snapshot.
 	mu sync.Mutex
 
@@ -50,7 +60,10 @@ type snapshot struct {
 
 	// files maps file URIs to their corresponding FileHandles.
 	// It may invalidated when a file's content changes.
-	files map[span.URI]source.FileHandle
+	files map[span.URI]source.VersionedFileHandle
+
+	// goFiles maps a parseKey to its parseGoHandle.
+	goFiles map[parseKey]*parseGoHandle
 
 	// packages maps a packageKey to a set of packageHandles to which that file belongs.
 	// It may be invalidated when a file's content changes.
@@ -63,6 +76,10 @@ type snapshot struct {
 	// when the view is created.
 	workspacePackages map[packageID]packagePath
 
+	// workspaceDirectories are the directories containing workspace packages.
+	// They are the view's root, as well as any replace targets.
+	workspaceDirectories map[span.URI]struct{}
+
 	// unloadableFiles keeps track of files that we've failed to load.
 	unloadableFiles map[span.URI]struct{}
 
@@ -71,13 +88,11 @@ type snapshot struct {
 	parseModHandles map[span.URI]*parseModHandle
 
 	// Preserve go.mod-related handles to avoid garbage-collecting the results
-	// of various calls to the go command.
-	//
-	// TODO(rstambler): If we end up with any more such handles, we should
-	// consider creating a struct for them.
-	modTidyHandle    *modTidyHandle
-	modWhyHandle     *modWhyHandle
-	modUpgradeHandle *modUpgradeHandle
+	// of various calls to the go command. The handles need not refer to only
+	// the view's go.mod file.
+	modTidyHandles    map[span.URI]*modTidyHandle
+	modUpgradeHandles map[span.URI]*modUpgradeHandle
+	modWhyHandles     map[span.URI]*modWhyHandle
 }
 
 type packageKey struct {
@@ -98,9 +113,22 @@ func (s *snapshot) View() source.View {
 	return s.view
 }
 
-// config returns the configuration used for the snapshot's interaction with the
-// go/packages API.
+func (s *snapshot) FileSet() *token.FileSet {
+	return s.view.session.cache.fset
+}
+
+// config returns a *packages.Config with the working directory set to the
+// view's root.
 func (s *snapshot) config(ctx context.Context) *packages.Config {
+	return s.configWithDir(ctx, s.view.root.Filename())
+}
+
+// configWithDir returns the configuration used for the snapshot's interaction
+// with the go/packages API. It uses the given working directory.
+// TODO(rstambler): go/packages requires that we do not provide overlays for
+// multiple modules in on config, so buildOverlay needs to filter overlays by
+// module.
+func (s *snapshot) configWithDir(ctx context.Context, dir string) *packages.Config {
 	s.view.optionsMu.Lock()
 	env, buildFlags := s.view.envLocked()
 	verboseOutput := s.view.options.VerboseOutput
@@ -108,7 +136,7 @@ func (s *snapshot) config(ctx context.Context) *packages.Config {
 
 	cfg := &packages.Config{
 		Context:    ctx,
-		Dir:        s.view.root.Filename(),
+		Dir:        dir,
 		Env:        append([]string{}, env...),
 		BuildFlags: append([]string{}, buildFlags...),
 		Mode: packages.NeedName |
@@ -140,83 +168,55 @@ func (s *snapshot) config(ctx context.Context) *packages.Config {
 }
 
 func (s *snapshot) RunGoCommandDirect(ctx context.Context, verb string, args []string) error {
-	cfg := s.config(ctx)
-	_, _, err := runGoCommand(ctx, cfg, nil, s.view.tmpMod, verb, args)
-	return err
-}
-
-func (s *snapshot) RunGoCommand(ctx context.Context, verb string, args []string) (*bytes.Buffer, error) {
-	cfg := s.config(ctx)
-	var pmh source.ParseModHandle
-	if s.view.tmpMod {
-		modFH, err := s.GetFile(ctx, s.view.modURI)
-		if err != nil {
-			return nil, err
-		}
-		pmh, err = s.ParseModHandle(ctx, modFH)
-		if err != nil {
-			return nil, err
-		}
-	}
-	_, stdout, err := runGoCommand(ctx, cfg, pmh, s.view.tmpMod, verb, args)
-	return stdout, err
-}
-
-func (s *snapshot) RunGoCommandPiped(ctx context.Context, verb string, args []string, stdout, stderr io.Writer) error {
-	cfg := s.config(ctx)
-	var pmh source.ParseModHandle
-	if s.view.tmpMod {
-		modFH, err := s.GetFile(ctx, s.view.modURI)
-		if err != nil {
-			return err
-		}
-		pmh, err = s.ParseModHandle(ctx, modFH)
-		if err != nil {
-			return err
-		}
-	}
-	_, inv, cleanup, err := goCommandInvocation(ctx, cfg, pmh, verb, args)
+	_, runner, inv, cleanup, err := s.goCommandInvocation(ctx, false, verb, args)
 	if err != nil {
 		return err
 	}
 	defer cleanup()
 
-	runner := packagesinternal.GetGoCmdRunner(cfg)
-	return runner.RunPiped(ctx, *inv, stdout, stderr)
+	_, err = runner.Run(ctx, *inv)
+	return err
 }
 
-// runGoCommand runs the given go command with the given config.
-// The given go.mod file is used to construct the temporary go.mod file, which
-// is then passed to the go command via the BuildFlags.
-// It assumes that modURI is only provided when the -modfile flag is enabled.
-func runGoCommand(ctx context.Context, cfg *packages.Config, pmh source.ParseModHandle, tmpMod bool, verb string, args []string) (span.URI, *bytes.Buffer, error) {
-	// Don't pass in the ParseModHandle if we are not using the -modfile flag.
-	var tmpPMH source.ParseModHandle
-	if tmpMod {
-		tmpPMH = pmh
-	}
-	tmpURI, inv, cleanup, err := goCommandInvocation(ctx, cfg, tmpPMH, verb, args)
+func (s *snapshot) RunGoCommand(ctx context.Context, verb string, args []string) (*bytes.Buffer, error) {
+	_, runner, inv, cleanup, err := s.goCommandInvocation(ctx, true, verb, args)
 	if err != nil {
-		return "", nil, err
+		return nil, err
 	}
 	defer cleanup()
 
-	runner := packagesinternal.GetGoCmdRunner(cfg)
-	stdout, err := runner.Run(ctx, *inv)
-	return tmpURI, stdout, err
+	return runner.Run(ctx, *inv)
+}
+
+func (s *snapshot) RunGoCommandPiped(ctx context.Context, verb string, args []string, stdout, stderr io.Writer) error {
+	_, runner, inv, cleanup, err := s.goCommandInvocation(ctx, true, verb, args)
+	if err != nil {
+		return err
+	}
+	defer cleanup()
+	return runner.RunPiped(ctx, *inv, stdout, stderr)
 }
 
 // Assumes that modURI is only provided when the -modfile flag is enabled.
-func goCommandInvocation(ctx context.Context, cfg *packages.Config, pmh source.ParseModHandle, verb string, args []string) (tmpURI span.URI, inv *gocommand.Invocation, cleanup func(), err error) {
+func (s *snapshot) goCommandInvocation(ctx context.Context, allowTempModfile bool, verb string, args []string) (tmpURI span.URI, runner *gocommand.Runner, inv *gocommand.Invocation, cleanup func(), err error) {
 	cleanup = func() {} // fallback
-	if pmh != nil {
-		tmpURI, cleanup, err = tempModFile(pmh.Mod(), pmh.Sum())
+	cfg := s.config(ctx)
+	if allowTempModfile && s.view.tmpMod {
+		modFH, err := s.GetFile(ctx, s.view.modURI)
 		if err != nil {
-			return "", nil, nil, err
+			return "", nil, nil, cleanup, err
+		}
+		// Use the go.sum if it happens to be available.
+		sumFH, _ := s.sumFH(ctx, modFH)
+
+		tmpURI, cleanup, err = tempModFile(modFH, sumFH)
+		if err != nil {
+			return "", nil, nil, cleanup, err
 		}
 		cfg.BuildFlags = append(cfg.BuildFlags, fmt.Sprintf("-modfile=%s", tmpURI.Filename()))
 	}
-	return tmpURI, &gocommand.Invocation{
+	runner = packagesinternal.GetGoCmdRunner(cfg)
+	return tmpURI, runner, &gocommand.Invocation{
 		Verb:       verb,
 		Args:       args,
 		Env:        cfg.Env,
@@ -244,7 +244,7 @@ func (s *snapshot) buildOverlay() map[string][]byte {
 	return overlays
 }
 
-func hashUnsavedOverlays(files map[span.URI]source.FileHandle) string {
+func hashUnsavedOverlays(files map[span.URI]source.VersionedFileHandle) string {
 	var unsaved []string
 	for uri, fh := range files {
 		if overlay, ok := fh.(*overlay); ok && !overlay.saved {
@@ -255,17 +255,13 @@ func hashUnsavedOverlays(files map[span.URI]source.FileHandle) string {
 	return hashContents([]byte(strings.Join(unsaved, "")))
 }
 
-func (s *snapshot) PackageHandles(ctx context.Context, fh source.FileHandle) ([]source.PackageHandle, error) {
-	if fh.Kind() != source.Go {
-		panic("called PackageHandles on a non-Go FileHandle")
-	}
-
-	ctx = event.Label(ctx, tag.URI.Of(fh.URI()))
+func (s *snapshot) PackagesForFile(ctx context.Context, uri span.URI, mode source.TypecheckMode) ([]source.Package, error) {
+	ctx = event.Label(ctx, tag.URI.Of(uri))
 
 	// Check if we should reload metadata for the file. We don't invalidate IDs
 	// (though we should), so the IDs will be a better source of truth than the
 	// metadata. If there are no IDs for the file, then we should also reload.
-	ids := s.getIDsForURI(fh.URI())
+	ids := s.getIDsForURI(uri)
 	reload := len(ids) == 0
 	for _, id := range ids {
 		// Reload package metadata if any of the metadata has missing
@@ -280,40 +276,47 @@ func (s *snapshot) PackageHandles(ctx context.Context, fh source.FileHandle) ([]
 		// calls to packages.Load. Determine what we should do instead.
 	}
 	if reload {
-		if err := s.load(ctx, fileURI(fh.URI())); err != nil {
+		if err := s.load(ctx, fileURI(uri)); err != nil {
 			return nil, err
 		}
 	}
 	// Get the list of IDs from the snapshot again, in case it has changed.
-	var phs []source.PackageHandle
-	for _, id := range s.getIDsForURI(fh.URI()) {
-		ph, err := s.packageHandle(ctx, id, source.ParseFull)
-		if err != nil {
-			return nil, err
+	var pkgs []source.Package
+	for _, id := range s.getIDsForURI(uri) {
+		var parseModes []source.ParseMode
+		switch mode {
+		case source.TypecheckAll:
+			if s.workspaceParseMode(id) == source.ParseFull {
+				parseModes = []source.ParseMode{source.ParseFull}
+			} else {
+				parseModes = []source.ParseMode{source.ParseExported, source.ParseFull}
+			}
+		case source.TypecheckFull:
+			parseModes = []source.ParseMode{source.ParseFull}
+		case source.TypecheckWorkspace:
+			parseModes = []source.ParseMode{s.workspaceParseMode(id)}
+		}
+
+		for _, parseMode := range parseModes {
+			pkg, err := s.checkedPackage(ctx, id, parseMode)
+			if err != nil {
+				return nil, err
+			}
+			pkgs = append(pkgs, pkg)
 		}
-		phs = append(phs, ph)
 	}
-	return phs, nil
+	return pkgs, nil
 }
 
-// packageHandle returns a PackageHandle for the given ID. It assumes that
-// the metadata for the given ID has already been loaded, but if the
-// PackageHandle has not been constructed, it will rebuild it.
-func (s *snapshot) packageHandle(ctx context.Context, id packageID, mode source.ParseMode) (*packageHandle, error) {
-	ph := s.getPackage(id, mode)
-	if ph != nil {
-		return ph, nil
+func (s *snapshot) checkedPackage(ctx context.Context, id packageID, mode source.ParseMode) (*pkg, error) {
+	ph, err := s.buildPackageHandle(ctx, id, mode)
+	if err != nil {
+		return nil, err
 	}
-	// Don't reload metadata in this function.
-	// Callers of this function must reload metadata themselves.
-	m := s.getMetadata(id)
-	if m == nil {
-		return nil, errors.Errorf("%s has no metadata", id)
-	}
-	return s.buildPackageHandle(ctx, m.id, mode)
+	return ph.check(ctx, s)
 }
 
-func (s *snapshot) GetReverseDependencies(ctx context.Context, id string) ([]source.PackageHandle, error) {
+func (s *snapshot) GetReverseDependencies(ctx context.Context, id string) ([]source.Package, error) {
 	if err := s.awaitLoaded(ctx); err != nil {
 		return nil, err
 	}
@@ -323,15 +326,15 @@ func (s *snapshot) GetReverseDependencies(ctx context.Context, id string) ([]sou
 	// Make sure to delete the original package ID from the map.
 	delete(ids, packageID(id))
 
-	var results []source.PackageHandle
+	var pkgs []source.Package
 	for id := range ids {
-		ph, err := s.packageHandle(ctx, id, source.ParseFull)
+		pkg, err := s.checkedPackage(ctx, id, s.workspaceParseMode(id))
 		if err != nil {
 			return nil, err
 		}
-		results = append(results, ph)
+		pkgs = append(pkgs, pkg)
 	}
-	return results, nil
+	return pkgs, nil
 }
 
 // transitiveReverseDependencies populates the uris map with file URIs
@@ -350,28 +353,44 @@ func (s *snapshot) transitiveReverseDependencies(id packageID, ids map[packageID
 	}
 }
 
-func (s *snapshot) getModHandle(uri span.URI) *parseModHandle {
+func (s *snapshot) getGoFile(key parseKey) *parseGoHandle {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+	return s.goFiles[key]
+}
+
+func (s *snapshot) addGoFile(key parseKey, pgh *parseGoHandle) *parseGoHandle {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+	if existing, ok := s.goFiles[key]; ok {
+		return existing
+	}
+	s.goFiles[key] = pgh
+	return pgh
+}
+
+func (s *snapshot) getParseModHandle(uri span.URI) *parseModHandle {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 	return s.parseModHandles[uri]
 }
 
-func (s *snapshot) getModWhyHandle() *modWhyHandle {
+func (s *snapshot) getModWhyHandle(uri span.URI) *modWhyHandle {
 	s.mu.Lock()
 	defer s.mu.Unlock()
-	return s.modWhyHandle
+	return s.modWhyHandles[uri]
 }
 
-func (s *snapshot) getModUpgradeHandle() *modUpgradeHandle {
+func (s *snapshot) getModUpgradeHandle(uri span.URI) *modUpgradeHandle {
 	s.mu.Lock()
 	defer s.mu.Unlock()
-	return s.modUpgradeHandle
+	return s.modUpgradeHandles[uri]
 }
 
-func (s *snapshot) getModTidyHandle() *modTidyHandle {
+func (s *snapshot) getModTidyHandle(uri span.URI) *modTidyHandle {
 	s.mu.Lock()
 	defer s.mu.Unlock()
-	return s.modTidyHandle
+	return s.modTidyHandles[uri]
 }
 
 func (s *snapshot) getImportedBy(id packageID) []packageID {
@@ -428,83 +447,77 @@ func (s *snapshot) workspacePackageIDs() (ids []packageID) {
 	return ids
 }
 
-func (s *snapshot) WorkspacePackages(ctx context.Context) ([]source.PackageHandle, error) {
-	if err := s.awaitLoaded(ctx); err != nil {
-		return nil, err
+func (s *snapshot) WorkspaceDirectories(ctx context.Context) []span.URI {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	var dirs []span.URI
+	for d := range s.workspaceDirectories {
+		dirs = append(dirs, d)
 	}
-	var results []source.PackageHandle
-	for _, pkgID := range s.workspacePackageIDs() {
-		ph, err := s.packageHandle(ctx, pkgID, source.ParseFull)
-		if err != nil {
-			return nil, err
-		}
-		results = append(results, ph)
-	}
-	return results, nil
+	return dirs
 }
 
-func (s *snapshot) KnownPackages(ctx context.Context) ([]source.PackageHandle, error) {
+func (s *snapshot) WorkspacePackages(ctx context.Context) ([]source.Package, error) {
 	if err := s.awaitLoaded(ctx); err != nil {
 		return nil, err
 	}
-	// Collect PackageHandles for all of the workspace packages first.
-	// They may need to be reloaded if their metadata has been invalidated.
-	wsPackages := make(map[packageID]bool)
-	s.mu.Lock()
-	for id := range s.workspacePackages {
-		wsPackages[id] = true
-	}
-	s.mu.Unlock()
-
-	var results []source.PackageHandle
-	for pkgID := range wsPackages {
-		ph, err := s.packageHandle(ctx, pkgID, source.ParseFull)
+	var pkgs []source.Package
+	for _, pkgID := range s.workspacePackageIDs() {
+		pkg, err := s.checkedPackage(ctx, pkgID, s.workspaceParseMode(pkgID))
 		if err != nil {
 			return nil, err
 		}
-		results = append(results, ph)
+		pkgs = append(pkgs, pkg)
+	}
+	return pkgs, nil
+}
+
+func (s *snapshot) KnownPackages(ctx context.Context) ([]source.Package, error) {
+	if err := s.awaitLoaded(ctx); err != nil {
+		return nil, err
 	}
 
-	// Once all workspace packages have been checked, the metadata will be up-to-date.
-	// Add all packages known in the workspace (that haven't already been added).
-	pkgIDs := make(map[packageID]bool)
+	// The WorkspaceSymbols implementation relies on this function returning
+	// workspace packages first.
+	ids := s.workspacePackageIDs()
 	s.mu.Lock()
 	for id := range s.metadata {
-		if !wsPackages[id] {
-			pkgIDs[id] = true
+		if _, ok := s.workspacePackages[id]; ok {
+			continue
 		}
+		ids = append(ids, id)
 	}
 	s.mu.Unlock()
 
-	for pkgID := range pkgIDs {
-		// Metadata for these packages should already be up-to-date,
-		// so just build the package handle directly (without a reload).
-		ph, err := s.buildPackageHandle(ctx, pkgID, source.ParseExported)
+	var pkgs []source.Package
+	for _, id := range ids {
+		pkg, err := s.checkedPackage(ctx, id, s.workspaceParseMode(id))
 		if err != nil {
 			return nil, err
 		}
-		results = append(results, ph)
+		pkgs = append(pkgs, pkg)
 	}
-	return results, nil
+	return pkgs, nil
 }
 
 func (s *snapshot) CachedImportPaths(ctx context.Context) (map[string]source.Package, error) {
 	// Don't reload workspace package metadata.
 	// This function is meant to only return currently cached information.
-	s.view.awaitInitialized(ctx)
+	s.view.AwaitInitialized(ctx)
 
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
 	results := map[string]source.Package{}
 	for _, ph := range s.packages {
-		cachedPkg, err := ph.cached()
+		cachedPkg, err := ph.cached(s.generation)
 		if err != nil {
 			continue
 		}
 		for importPath, newPkg := range cachedPkg.imports {
 			if oldPkg, ok := results[string(importPath)]; ok {
-				// Using the same trick as NarrowestPackageHandle, prefer non-variants.
+				// Using the same trick as NarrowestPackage, prefer non-variants.
 				if len(newPkg.compiledGoFiles) < len(oldPkg.(*pkg).compiledGoFiles) {
 					results[string(importPath)] = newPkg
 				}
@@ -548,7 +561,7 @@ func (s *snapshot) addActionHandle(ah *actionHandle) *actionHandle {
 	key := actionKey{
 		analyzer: ah.analyzer,
 		pkg: packageKey{
-			id:   ah.pkg.id,
+			id:   ah.pkg.m.id,
 			mode: ah.pkg.mode,
 		},
 	}
@@ -616,7 +629,7 @@ func (s *snapshot) isWorkspacePackage(id packageID) (packagePath, bool) {
 	return scope, ok
 }
 
-func (s *snapshot) FindFile(uri span.URI) source.FileHandle {
+func (s *snapshot) FindFile(uri span.URI) source.VersionedFileHandle {
 	f, err := s.view.getFile(uri)
 	if err != nil {
 		return nil
@@ -630,7 +643,7 @@ func (s *snapshot) FindFile(uri span.URI) source.FileHandle {
 
 // GetFile returns a File for the given URI. It will always succeed because it
 // adds the file to the managed set if needed.
-func (s *snapshot) GetFile(ctx context.Context, uri span.URI) (source.FileHandle, error) {
+func (s *snapshot) GetFile(ctx context.Context, uri span.URI) (source.VersionedFileHandle, error) {
 	f, err := s.view.getFile(uri)
 	if err != nil {
 		return nil, err
@@ -647,8 +660,9 @@ func (s *snapshot) GetFile(ctx context.Context, uri span.URI) (source.FileHandle
 	if err != nil {
 		return nil, err
 	}
-	s.files[f.URI()] = fh
-	return fh, nil
+	closed := &closedFile{fh}
+	s.files[f.URI()] = closed
+	return closed, nil
 }
 
 func (s *snapshot) IsOpen(uri span.URI) bool {
@@ -669,7 +683,7 @@ func (s *snapshot) IsSaved(uri span.URI) bool {
 
 func (s *snapshot) awaitLoaded(ctx context.Context) error {
 	// Do not return results until the snapshot's view has been initialized.
-	s.view.awaitInitialized(ctx)
+	s.view.AwaitInitialized(ctx)
 
 	if err := s.reloadWorkspace(ctx); err != nil {
 		return err
@@ -698,21 +712,25 @@ func (s *snapshot) reloadWorkspace(ctx context.Context) error {
 
 	// See which of the workspace packages are missing metadata.
 	s.mu.Lock()
-	var pkgPaths []interface{}
+	pkgPathSet := map[packagePath]struct{}{}
 	for id, pkgPath := range s.workspacePackages {
 		// Don't try to reload "command-line-arguments" directly.
 		if pkgPath == "command-line-arguments" {
 			continue
 		}
 		if s.metadata[id] == nil {
-			pkgPaths = append(pkgPaths, pkgPath)
+			pkgPathSet[pkgPath] = struct{}{}
 		}
 	}
 	s.mu.Unlock()
 
-	if len(pkgPaths) == 0 {
+	if len(pkgPathSet) == 0 {
 		return nil
 	}
+	var pkgPaths []interface{}
+	for pkgPath := range pkgPathSet {
+		pkgPaths = append(pkgPaths, pkgPath)
+	}
 	return s.load(ctx, pkgPaths...)
 }
 
@@ -790,25 +808,38 @@ func contains(views []*View, view *View) bool {
 	return false
 }
 
-func (s *snapshot) clone(ctx context.Context, withoutURIs map[span.URI]source.FileHandle, forceReloadMetadata bool) *snapshot {
+func generationName(v *View, snapshotID uint64) string {
+	return fmt.Sprintf("v%v/%v", v.id, snapshotID)
+}
+
+func (s *snapshot) clone(ctx context.Context, withoutURIs map[span.URI]source.VersionedFileHandle, forceReloadMetadata bool) *snapshot {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
+	newGen := s.view.session.cache.store.Generation(generationName(s.view, s.id+1))
 	result := &snapshot{
-		id:                s.id + 1,
-		view:              s.view,
-		ids:               make(map[span.URI][]packageID),
-		importedBy:        make(map[packageID][]packageID),
-		metadata:          make(map[packageID]*metadata),
-		packages:          make(map[packageKey]*packageHandle),
-		actions:           make(map[actionKey]*actionHandle),
-		files:             make(map[span.URI]source.FileHandle),
-		workspacePackages: make(map[packageID]packagePath),
-		unloadableFiles:   make(map[span.URI]struct{}),
-		parseModHandles:   make(map[span.URI]*parseModHandle),
-		modTidyHandle:     s.modTidyHandle,
-		modUpgradeHandle:  s.modUpgradeHandle,
-		modWhyHandle:      s.modWhyHandle,
+		id:                   s.id + 1,
+		generation:           newGen,
+		view:                 s.view,
+		builtin:              s.builtin,
+		ids:                  make(map[span.URI][]packageID),
+		importedBy:           make(map[packageID][]packageID),
+		metadata:             make(map[packageID]*metadata),
+		packages:             make(map[packageKey]*packageHandle),
+		actions:              make(map[actionKey]*actionHandle),
+		files:                make(map[span.URI]source.VersionedFileHandle),
+		goFiles:              make(map[parseKey]*parseGoHandle),
+		workspaceDirectories: make(map[span.URI]struct{}),
+		workspacePackages:    make(map[packageID]packagePath),
+		unloadableFiles:      make(map[span.URI]struct{}),
+		parseModHandles:      make(map[span.URI]*parseModHandle),
+		modTidyHandles:       make(map[span.URI]*modTidyHandle),
+		modUpgradeHandles:    make(map[span.URI]*modUpgradeHandle),
+		modWhyHandles:        make(map[span.URI]*modWhyHandle),
+	}
+
+	if s.builtin != nil {
+		newGen.Inherit(s.builtin.handle)
 	}
 
 	// Copy all of the FileHandles.
@@ -821,8 +852,43 @@ func (s *snapshot) clone(ctx context.Context, withoutURIs map[span.URI]source.Fi
 	}
 	// Copy all of the modHandles.
 	for k, v := range s.parseModHandles {
+		newGen.Inherit(v.handle)
 		result.parseModHandles[k] = v
 	}
+	// Copy all of the workspace directories. They may be reset later.
+	for k, v := range s.workspaceDirectories {
+		result.workspaceDirectories[k] = v
+	}
+
+	for k, v := range s.goFiles {
+		if _, ok := withoutURIs[k.file.URI]; ok {
+			continue
+		}
+		newGen.Inherit(v.handle)
+		newGen.Inherit(v.astCacheHandle)
+		result.goFiles[k] = v
+	}
+
+	// Copy all of the go.mod-related handles. They may be invalidated later,
+	// so we inherit them at the end of the function.
+	for k, v := range s.modTidyHandles {
+		if _, ok := withoutURIs[k]; ok {
+			continue
+		}
+		result.modTidyHandles[k] = v
+	}
+	for k, v := range s.modUpgradeHandles {
+		if _, ok := withoutURIs[k]; ok {
+			continue
+		}
+		result.modUpgradeHandles[k] = v
+	}
+	for k, v := range s.modWhyHandles {
+		if _, ok := withoutURIs[k]; ok {
+			continue
+		}
+		result.modWhyHandles[k] = v
+	}
 
 	// transitiveIDs keeps track of transitive reverse dependencies.
 	// If an ID is present in the map, invalidate its types.
@@ -841,24 +907,40 @@ func (s *snapshot) clone(ctx context.Context, withoutURIs map[span.URI]source.Fi
 
 		// Check if the file's package name or imports have changed,
 		// and if so, invalidate this file's packages' metadata.
-		invalidateMetadata := forceReloadMetadata || s.shouldInvalidateMetadata(ctx, originalFH, currentFH)
+		invalidateMetadata := forceReloadMetadata || s.shouldInvalidateMetadata(ctx, result, originalFH, currentFH)
 
 		// Invalidate the previous modTidyHandle if any of the files have been
 		// saved or if any of the metadata has been invalidated.
 		if invalidateMetadata || fileWasSaved(originalFH, currentFH) {
-			result.modTidyHandle = nil
-			result.modUpgradeHandle = nil
-			result.modWhyHandle = nil
+			// TODO(rstambler): Only delete mod handles for which the
+			// withoutURI is relevant.
+			for k := range s.modTidyHandles {
+				delete(result.modTidyHandles, k)
+			}
+			for k := range s.modUpgradeHandles {
+				delete(result.modUpgradeHandles, k)
+			}
+			for k := range s.modWhyHandles {
+				delete(result.modWhyHandles, k)
+			}
 		}
 		if currentFH.Kind() == source.Mod {
-			// If the view's go.mod file's contents have changed, invalidate the metadata
-			// for all of the packages in the workspace.
+			// If the view's go.mod file's contents have changed, invalidate the
+			// metadata for every known package in the snapshot.
 			if invalidateMetadata {
-				for id := range s.workspacePackages {
-					directIDs[id] = struct{}{}
+				for k := range s.packages {
+					directIDs[k.id] = struct{}{}
 				}
 			}
+
 			delete(result.parseModHandles, withoutURI)
+
+			if currentFH.URI() == s.view.modURI {
+				// The go.mod's replace directives may have changed. We may
+				// need to update our set of workspace directories. Use the new
+				// snapshot, as it can be locked without causing issues.
+				result.workspaceDirectories = result.findWorkspaceDirectories(ctx, currentFH)
+			}
 		}
 
 		// If this is a file we don't yet know about,
@@ -915,6 +997,7 @@ func (s *snapshot) clone(ctx context.Context, withoutURIs map[span.URI]source.Fi
 		if _, ok := transitiveIDs[k.id]; ok {
 			continue
 		}
+		newGen.Inherit(v.handle)
 		result.packages[k] = v
 	}
 	// Copy the package analysis information.
@@ -922,6 +1005,7 @@ func (s *snapshot) clone(ctx context.Context, withoutURIs map[span.URI]source.Fi
 		if _, ok := transitiveIDs[k.pkg.id]; ok {
 			continue
 		}
+		newGen.Inherit(v.handle)
 		result.actions[k] = v
 	}
 	// Copy the package metadata. We only need to invalidate packages directly
@@ -968,6 +1052,18 @@ copyIDs:
 
 		result.workspacePackages[id] = pkgPath
 	}
+
+	// Inherit all of the go.mod-related handles.
+	for _, v := range s.modTidyHandles {
+		newGen.Inherit(v.handle)
+	}
+	for _, v := range s.modUpgradeHandles {
+		newGen.Inherit(v.handle)
+	}
+	for _, v := range s.modWhyHandles {
+		newGen.Inherit(v.handle)
+	}
+
 	// Don't bother copying the importedBy graph,
 	// as it changes each time we update metadata.
 	return result
@@ -991,43 +1087,145 @@ func fileWasSaved(originalFH, currentFH source.FileHandle) bool {
 
 // shouldInvalidateMetadata reparses a file's package and import declarations to
 // determine if the file requires a metadata reload.
-func (s *snapshot) shouldInvalidateMetadata(ctx context.Context, originalFH, currentFH source.FileHandle) bool {
+func (s *snapshot) shouldInvalidateMetadata(ctx context.Context, newSnapshot *snapshot, originalFH, currentFH source.FileHandle) bool {
 	if originalFH == nil {
 		return currentFH.Kind() == source.Go
 	}
 	// If the file hasn't changed, there's no need to reload.
-	if originalFH.Identity().String() == currentFH.Identity().String() {
+	if originalFH.FileIdentity() == currentFH.FileIdentity() {
 		return false
 	}
 	// If a go.mod file's contents have changed, always invalidate metadata.
 	if kind := originalFH.Kind(); kind == source.Mod {
 		return originalFH.URI() == s.view.modURI
 	}
-	// Get the original and current parsed files in order to check package name and imports.
-	original, _, _, _, originalErr := s.view.session.cache.ParseGoHandle(ctx, originalFH, source.ParseHeader).Parse(ctx)
-	current, _, _, _, currentErr := s.view.session.cache.ParseGoHandle(ctx, currentFH, source.ParseHeader).Parse(ctx)
+	// Get the original and current parsed files in order to check package name
+	// and imports. Use the new snapshot to parse to avoid modifying the
+	// current snapshot.
+	original, originalErr := newSnapshot.ParseGo(ctx, originalFH, source.ParseHeader)
+	current, currentErr := newSnapshot.ParseGo(ctx, currentFH, source.ParseHeader)
 	if originalErr != nil || currentErr != nil {
 		return (originalErr == nil) != (currentErr == nil)
 	}
 	// Check if the package's metadata has changed. The cases handled are:
 	//    1. A package's name has changed
 	//    2. A file's imports have changed
-	if original.Name.Name != current.Name.Name {
-		return true
-	}
-	// If the package's imports have increased, definitely re-run `go list`.
-	if len(original.Imports) < len(current.Imports) {
+	if original.File.Name.Name != current.File.Name.Name {
 		return true
 	}
 	importSet := make(map[string]struct{})
-	for _, importSpec := range original.Imports {
+	for _, importSpec := range original.File.Imports {
 		importSet[importSpec.Path.Value] = struct{}{}
 	}
 	// If any of the current imports were not in the original imports.
-	for _, importSpec := range current.Imports {
-		if _, ok := importSet[importSpec.Path.Value]; !ok {
-			return true
+	for _, importSpec := range current.File.Imports {
+		if _, ok := importSet[importSpec.Path.Value]; ok {
+			continue
 		}
+		// If the import path is obviously not valid, we can skip reloading
+		// metadata. For now, valid means properly quoted and without a
+		// terminal slash.
+		path, err := strconv.Unquote(importSpec.Path.Value)
+		if err != nil {
+			continue
+		}
+		if path == "" {
+			continue
+		}
+		if path[len(path)-1] == '/' {
+			continue
+		}
+		return true
 	}
 	return false
 }
+
+// findWorkspaceDirectoriesLocked returns all of the directories that are
+// considered to be part of the view's workspace. For GOPATH workspaces, this
+// is just the view's root. For modules-based workspaces, this is the module
+// root and any replace targets. It also returns the parseModHandle for the
+// view's go.mod file if it has one.
+//
+// It assumes that the file handle is the view's go.mod file, if it has one.
+// The caller need not be holding the snapshot's mutex, but it might be.
+func (s *snapshot) findWorkspaceDirectories(ctx context.Context, modFH source.FileHandle) map[span.URI]struct{} {
+	m := map[span.URI]struct{}{
+		s.view.root: {},
+	}
+	// If the view does not have a go.mod file, only the root directory
+	// is known. In GOPATH mode, we should really watch the entire GOPATH,
+	// but that's too expensive.
+	modURI := s.view.modURI
+	if modURI == "" {
+		return m
+	}
+	if modFH == nil {
+		return m
+	}
+	// Ignore parse errors. An invalid go.mod is not fatal.
+	mod, err := s.ParseMod(ctx, modFH)
+	if err != nil {
+		return m
+	}
+	for _, r := range mod.File.Replace {
+		// We may be replacing a module with a different version, not a path
+		// on disk.
+		if r.New.Version != "" {
+			continue
+		}
+		uri := span.URIFromPath(r.New.Path)
+		m[uri] = struct{}{}
+	}
+	return m
+}
+
+func (s *snapshot) BuiltinPackage(ctx context.Context) (*source.BuiltinPackage, error) {
+	s.view.AwaitInitialized(ctx)
+
+	if s.builtin == nil {
+		return nil, errors.Errorf("no builtin package for view %s", s.view.name)
+	}
+	d, err := s.builtin.handle.Get(ctx, s.generation, s)
+	if err != nil {
+		return nil, err
+	}
+	data := d.(*builtinPackageData)
+	return data.parsed, data.err
+}
+
+func (s *snapshot) buildBuiltinPackage(ctx context.Context, goFiles []string) error {
+	if len(goFiles) != 1 {
+		return errors.Errorf("only expected 1 file, got %v", len(goFiles))
+	}
+	uri := span.URIFromPath(goFiles[0])
+
+	// Get the FileHandle through the cache to avoid adding it to the snapshot
+	// and to get the file content from disk.
+	fh, err := s.view.session.cache.getFile(ctx, uri)
+	if err != nil {
+		return err
+	}
+	h := s.generation.Bind(fh.FileIdentity(), func(ctx context.Context, arg memoize.Arg) interface{} {
+		snapshot := arg.(*snapshot)
+
+		pgh := snapshot.parseGoHandle(ctx, fh, source.ParseFull)
+		pgf, _, err := snapshot.parseGo(ctx, pgh)
+		if err != nil {
+			return &builtinPackageData{err: err}
+		}
+		pkg, err := ast.NewPackage(snapshot.view.session.cache.fset, map[string]*ast.File{
+			pgf.URI.Filename(): pgf.File,
+		}, nil, nil)
+		if err != nil {
+			return &builtinPackageData{err: err}
+		}
+		return &builtinPackageData{
+			parsed: &source.BuiltinPackage{
+				ParsedFile: pgf,
+				Package:    pkg,
+			},
+		}
+	})
+	s.builtin = &builtinPackageHandle{handle: h}
+	return nil
+}
