@@ -9,13 +9,14 @@ import (
 	"context"
 	"encoding/json"
 	"fmt"
-	"go/ast"
 	"io"
 	"io/ioutil"
 	"os"
+	"os/exec"
 	"path"
 	"path/filepath"
 	"reflect"
+	"sort"
 	"strings"
 	"sync"
 	"time"
@@ -115,9 +116,6 @@ type View struct {
 	initializeOnce     *sync.Once
 	initializedErr     error
 
-	// builtin pins the AST and package for builtin.go in memory.
-	builtin *builtinPackageHandle
-
 	// True if the view is either in GOPATH, a module, or some other
 	// non go command build system.
 	hasValidBuildConfiguration bool
@@ -129,9 +127,10 @@ type View struct {
 	// Only possible with Go versions 1.14 and above.
 	tmpMod bool
 
-	// goCommand indicates if the user is using the go command or some other
-	// build system.
-	goCommand bool
+	// hasGopackagesDriver is true if the user has a value set for the
+	// GOPACKAGESDRIVER environment variable or a gopackagesdriver binary on
+	// their machine.
+	hasGopackagesDriver bool
 
 	// `go env` variables that need to be tracked by gopls.
 	gocache, gomodcache, gopath, goprivate string
@@ -143,23 +142,11 @@ type View struct {
 
 type builtinPackageHandle struct {
 	handle *memoize.Handle
-	file   source.ParseGoHandle
 }
 
 type builtinPackageData struct {
-	memoize.NoCopy
-
-	pkg *ast.Package
-	pgh *parseGoHandle
-	err error
-}
-
-func (d *builtinPackageData) Package() *ast.Package {
-	return d.pkg
-}
-
-func (d *builtinPackageData) ParseGoHandle() source.ParseGoHandle {
-	return d.pgh
+	parsed *source.BuiltinPackage
+	err    error
 }
 
 // fileBase holds the common functionality for all files.
@@ -234,10 +221,10 @@ func tempModFile(modFh, sumFH source.FileHandle) (tmpURI span.URI, cleanup func(
 	if sumFH != nil {
 		sumContents, err := sumFH.Read()
 		if err != nil {
-			return "", nil, err
+			return "", cleanup, err
 		}
 		if err := ioutil.WriteFile(tmpSumName, sumContents, 0655); err != nil {
-			return "", nil, err
+			return "", cleanup, err
 		}
 	}
 
@@ -266,10 +253,30 @@ func (v *View) Options() source.Options {
 
 func minorOptionsChange(a, b source.Options) bool {
 	// Check if any of the settings that modify our understanding of files have been changed
-	if !reflect.DeepEqual(a.Env, b.Env) {
+	mapEnv := func(env []string) map[string]string {
+		m := make(map[string]string, len(env))
+		for _, x := range env {
+			split := strings.SplitN(x, "=", 2)
+			if len(split) != 2 {
+				continue
+			}
+			m[split[0]] = split[1]
+
+		}
+		return m
+	}
+	aEnv := mapEnv(a.Env)
+	bEnv := mapEnv(b.Env)
+	if !reflect.DeepEqual(aEnv, bEnv) {
 		return false
 	}
-	if !reflect.DeepEqual(a.BuildFlags, b.BuildFlags) {
+	aBuildFlags := make([]string, len(a.BuildFlags))
+	bBuildFlags := make([]string, len(b.BuildFlags))
+	copy(aBuildFlags, a.BuildFlags)
+	copy(bBuildFlags, b.BuildFlags)
+	sort.Strings(aBuildFlags)
+	sort.Strings(bBuildFlags)
+	if !reflect.DeepEqual(aBuildFlags, bBuildFlags) {
 		return false
 	}
 	// the rest of the options are benign
@@ -285,76 +292,17 @@ func (v *View) SetOptions(ctx context.Context, options source.Options) (source.V
 		return v, nil
 	}
 	v.optionsMu.Unlock()
-	newView, _, err := v.session.updateView(ctx, v, options)
+	newView, err := v.session.updateView(ctx, v, options)
 	return newView, err
 }
 
-func (v *View) Rebuild(ctx context.Context) (source.Snapshot, error) {
-	_, snapshot, err := v.session.updateView(ctx, v, v.Options())
-	return snapshot, err
-}
-
-func (v *View) BuiltinPackage(ctx context.Context) (source.BuiltinPackage, error) {
-	v.awaitInitialized(ctx)
-
-	if v.builtin == nil {
-		return nil, errors.Errorf("no builtin package for view %s", v.name)
-	}
-	data, err := v.builtin.handle.Get(ctx)
+func (v *View) Rebuild(ctx context.Context) (source.Snapshot, func(), error) {
+	newView, err := v.session.updateView(ctx, v, v.Options())
 	if err != nil {
-		return nil, err
+		return nil, func() {}, err
 	}
-	if data == nil {
-		return nil, errors.Errorf("unexpected nil builtin package")
-	}
-	d, ok := data.(*builtinPackageData)
-	if !ok {
-		return nil, errors.Errorf("unexpected type %T", data)
-	}
-	if d.err != nil {
-		return nil, d.err
-	}
-	if d.pkg == nil || d.pkg.Scope == nil {
-		return nil, errors.Errorf("no builtin package")
-	}
-	return d, nil
-}
-
-func (v *View) buildBuiltinPackage(ctx context.Context, goFiles []string) error {
-	if len(goFiles) != 1 {
-		return errors.Errorf("only expected 1 file, got %v", len(goFiles))
-	}
-	uri := span.URIFromPath(goFiles[0])
-
-	// Get the FileHandle through the cache to avoid adding it to the snapshot
-	// and to get the file content from disk.
-	fh, err := v.session.cache.getFile(ctx, uri)
-	if err != nil {
-		return err
-	}
-	pgh := v.session.cache.parseGoHandle(ctx, fh, source.ParseFull)
-	fset := v.session.cache.fset
-	h := v.session.cache.store.Bind(fh.Identity(), func(ctx context.Context) interface{} {
-		file, _, _, _, err := pgh.Parse(ctx)
-		if err != nil {
-			return &builtinPackageData{err: err}
-		}
-		pkg, err := ast.NewPackage(fset, map[string]*ast.File{
-			pgh.File().URI().Filename(): file,
-		}, nil, nil)
-		if err != nil {
-			return &builtinPackageData{err: err}
-		}
-		return &builtinPackageData{
-			pgh: pgh,
-			pkg: pkg,
-		}
-	})
-	v.builtin = &builtinPackageHandle{
-		handle: h,
-		file:   pgh,
-	}
-	return nil
+	snapshot, release := newView.Snapshot(ctx)
+	return snapshot, release, nil
 }
 
 func (v *View) WriteEnv(ctx context.Context, w io.Writer) error {
@@ -399,7 +347,7 @@ func (v *View) RunProcessEnvFunc(ctx context.Context, fn func(*imports.Options)
 		if err != nil {
 			return err
 		}
-		modFileIdentifier = modFH.Identity().Identifier
+		modFileIdentifier = modFH.FileIdentity().Hash
 	}
 	if v.sumURI != "" {
 		sumFH, err = v.session.cache.getFile(ctx, v.sumURI)
@@ -545,36 +493,6 @@ func basename(filename string) string {
 	return strings.ToLower(filepath.Base(filename))
 }
 
-func (v *View) WorkspaceDirectories(ctx context.Context) ([]string, error) {
-	// If the view does not have a go.mod file, only the root directory
-	// is known. In GOPATH mode, we should really watch the entire GOPATH,
-	// but that's probably too expensive.
-	// TODO(rstambler): Figure out a better approach in the future.
-	if v.modURI == "" {
-		return []string{v.root.Filename()}, nil
-	}
-	// Anything inside of the module root is known.
-	dirs := []string{filepath.Dir(v.modURI.Filename())}
-
-	// Keep track of any directories mentioned in replace targets.
-	fh, err := v.session.GetFile(ctx, v.modURI)
-	if err != nil {
-		return nil, err
-	}
-	pmh, err := v.Snapshot().ParseModHandle(ctx, fh)
-	if err != nil {
-		return nil, err
-	}
-	parsed, _, _, err := pmh.Parse(ctx)
-	if err != nil {
-		return nil, err
-	}
-	for _, replace := range parsed.Replace {
-		dirs = append(dirs, replace.New.Path)
-	}
-	return dirs, nil
-}
-
 func (v *View) relevantChange(c source.FileModification) bool {
 	// If the file is known to the view, the change is relevant.
 	known := v.knownFile(c.URI)
@@ -660,11 +578,14 @@ func (v *View) shutdown(ctx context.Context) {
 	v.initCancelFirstAttempt()
 
 	v.mu.Lock()
-	defer v.mu.Unlock()
 	if v.cancel != nil {
 		v.cancel()
 		v.cancel = nil
 	}
+	v.mu.Unlock()
+	v.snapshotMu.Lock()
+	go v.snapshot.generation.Destroy()
+	v.snapshotMu.Unlock()
 }
 
 func (v *View) BackgroundContext() context.Context {
@@ -709,8 +630,9 @@ func checkIgnored(suffix string) bool {
 	return false
 }
 
-func (v *View) Snapshot() source.Snapshot {
-	return v.getSnapshot()
+func (v *View) Snapshot(ctx context.Context) (source.Snapshot, func()) {
+	s := v.getSnapshot()
+	return s, s.generation.Acquire(ctx)
 }
 
 func (v *View) getSnapshot() *snapshot {
@@ -753,7 +675,8 @@ func (v *View) initialize(ctx context.Context, s *snapshot, firstAttempt bool) {
 	})
 }
 
-func (v *View) awaitInitialized(ctx context.Context) {
+// AwaitInitialized waits until a view is initialized
+func (v *View) AwaitInitialized(ctx context.Context) {
 	select {
 	case <-ctx.Done():
 		return
@@ -767,7 +690,7 @@ func (v *View) awaitInitialized(ctx context.Context) {
 // invalidateContent invalidates the content of a Go file,
 // including any position and type information that depends on it.
 // It returns true if we were already tracking the given file, false otherwise.
-func (v *View) invalidateContent(ctx context.Context, uris map[span.URI]source.FileHandle, forceReloadMetadata bool) source.Snapshot {
+func (v *View) invalidateContent(ctx context.Context, uris map[span.URI]source.VersionedFileHandle, forceReloadMetadata bool) (source.Snapshot, func()) {
 	// Detach the context so that content invalidation cannot be canceled.
 	ctx = xcontext.Detach(ctx)
 
@@ -776,14 +699,17 @@ func (v *View) invalidateContent(ctx context.Context, uris map[span.URI]source.F
 	v.cancelBackground()
 
 	// Do not clone a snapshot until its view has finished initializing.
-	v.awaitInitialized(ctx)
+	v.AwaitInitialized(ctx)
 
 	// This should be the only time we hold the view's snapshot lock for any period of time.
 	v.snapshotMu.Lock()
 	defer v.snapshotMu.Unlock()
 
-	v.snapshot = v.snapshot.clone(ctx, uris, forceReloadMetadata)
-	return v.snapshot
+	oldSnapshot := v.snapshot
+	v.snapshot = oldSnapshot.clone(ctx, uris, forceReloadMetadata)
+	go oldSnapshot.generation.Destroy()
+
+	return v.snapshot, v.snapshot.generation.Acquire(ctx)
 }
 
 func (v *View) cancelBackground() {
@@ -810,12 +736,12 @@ func (v *View) maybeReinitialize() {
 	v.initializeOnce = &once
 }
 
-func (v *View) setBuildInformation(ctx context.Context, folder span.URI, env []string, modfileFlagEnabled bool) error {
+func (v *View) setBuildInformation(ctx context.Context, folder span.URI, options source.Options) error {
 	if err := checkPathCase(folder.Filename()); err != nil {
-		return fmt.Errorf("invalid workspace configuration: %w", err)
+		return errors.Errorf("invalid workspace configuration: %w", err)
 	}
 	// Make sure to get the `go env` before continuing with initialization.
-	modFile, err := v.setGoEnv(ctx, env)
+	modFile, err := v.setGoEnv(ctx, options.Env)
 	if err != nil {
 		return err
 	}
@@ -829,8 +755,7 @@ func (v *View) setBuildInformation(ctx context.Context, folder span.URI, env []s
 		v.sumURI = span.URIFromPath(sumFilename)
 	}
 
-	v.root = v.folder
-	if v.modURI != "" {
+	if options.ExpandWorkspaceToModule && v.modURI != "" {
 		v.root = span.URIFromPath(filepath.Dir(v.modURI.Filename()))
 	}
 
@@ -839,7 +764,7 @@ func (v *View) setBuildInformation(ctx context.Context, folder span.URI, env []s
 	v.setBuildConfiguration()
 
 	// The user has disabled the use of the -modfile flag or has no go.mod file.
-	if !modfileFlagEnabled || v.modURI == "" {
+	if !options.TempModfile || v.modURI == "" {
 		return nil
 	}
 	if modfileFlag, err := v.modfileFlagExists(ctx, v.Options().Env); err != nil {
@@ -861,9 +786,9 @@ func (v *View) setBuildConfiguration() (isValid bool) {
 	defer func() {
 		v.hasValidBuildConfiguration = isValid
 	}()
-	// Since we only really understand the `go` command, if the user is not
-	// using the go command, assume that their configuration is valid.
-	if !v.goCommand {
+	// Since we only really understand the `go` command, if the user has a
+	// different GOPACKAGESDRIVER, assume that their configuration is valid.
+	if v.hasGopackagesDriver {
 		return true
 	}
 	// Check if the user is working within a module.
@@ -929,7 +854,16 @@ func (v *View) setGoEnv(ctx context.Context, configEnv []string) (string, error)
 
 	// The value of GOPACKAGESDRIVER is not returned through the go command.
 	gopackagesdriver := os.Getenv("GOPACKAGESDRIVER")
-	v.goCommand = gopackagesdriver == "" || gopackagesdriver == "off"
+	for _, s := range configEnv {
+		split := strings.SplitN(s, "=", 2)
+		if split[0] == "GOPACKAGESDRIVER" {
+			gopackagesdriver = split[1]
+		}
+	}
+	// A user may also have a gopackagesdriver binary on their machine, which
+	// works the same way as setting GOPACKAGESDRIVER.
+	tool, _ := exec.LookPath("gopackagesdriver")
+	v.hasGopackagesDriver = gopackagesdriver != "off" && (gopackagesdriver != "" || tool != "")
 	return gomod, nil
 }
 
