@@ -20,7 +20,8 @@ import (
 )
 
 func (s *Server) codeAction(ctx context.Context, params *protocol.CodeActionParams) ([]protocol.CodeAction, error) {
-	snapshot, fh, ok, err := s.beginFileRequest(ctx, params.TextDocument.URI, source.UnknownKind)
+	snapshot, fh, ok, release, err := s.beginFileRequest(ctx, params.TextDocument.URI, source.UnknownKind)
+	defer release()
 	if !ok {
 		return nil, err
 	}
@@ -118,16 +119,16 @@ func (s *Server) codeAction(ctx context.Context, params *protocol.CodeActionPara
 		if ctx.Err() != nil {
 			return nil, ctx.Err()
 		}
-		phs, err := snapshot.PackageHandles(ctx, fh)
+		pkgs, err := snapshot.PackagesForFile(ctx, fh.URI(), source.TypecheckFull)
 		if err != nil {
 			return nil, err
 		}
-		ph, err := source.WidestPackageHandle(phs)
+		pkg, err := source.WidestPackage(pkgs)
 		if err != nil {
 			return nil, err
 		}
 		if (wanted[protocol.QuickFix] || wanted[protocol.SourceFixAll]) && len(diagnostics) > 0 {
-			analysisQuickFixes, highConfidenceEdits, err := analysisFixes(ctx, snapshot, ph, diagnostics)
+			analysisQuickFixes, highConfidenceEdits, err := analysisFixes(ctx, snapshot, pkg, diagnostics)
 			if err != nil {
 				return nil, err
 			}
@@ -159,14 +160,14 @@ func (s *Server) codeAction(ctx context.Context, params *protocol.CodeActionPara
 		}
 		// Add any suggestions that do not necessarily fix any diagnostics.
 		if wanted[protocol.RefactorRewrite] {
-			fixes, err := convenienceFixes(ctx, snapshot, ph, uri, params.Range)
+			fixes, err := convenienceFixes(ctx, snapshot, pkg, uri, params.Range)
 			if err != nil {
 				return nil, err
 			}
 			codeActions = append(codeActions, fixes...)
 		}
 		if wanted[protocol.RefactorExtract] {
-			fixes, err := extractionFixes(ctx, snapshot, ph, uri, params.Range)
+			fixes, err := extractionFixes(ctx, snapshot, pkg, uri, params.Range)
 			if err != nil {
 				return nil, err
 			}
@@ -237,7 +238,7 @@ func importDiagnostics(fix *imports.ImportFix, diagnostics []protocol.Diagnostic
 	return results
 }
 
-func analysisFixes(ctx context.Context, snapshot source.Snapshot, ph source.PackageHandle, diagnostics []protocol.Diagnostic) ([]protocol.CodeAction, []protocol.TextDocumentEdit, error) {
+func analysisFixes(ctx context.Context, snapshot source.Snapshot, pkg source.Package, diagnostics []protocol.Diagnostic) ([]protocol.CodeAction, []protocol.TextDocumentEdit, error) {
 	if len(diagnostics) == 0 {
 		return nil, nil, nil
 	}
@@ -246,7 +247,7 @@ func analysisFixes(ctx context.Context, snapshot source.Snapshot, ph source.Pack
 		sourceFixAllEdits []protocol.TextDocumentEdit
 	)
 	for _, diag := range diagnostics {
-		srcErr, analyzer, ok := findSourceError(ctx, snapshot, ph.ID(), diag)
+		srcErr, analyzer, ok := findSourceError(ctx, snapshot, pkg.ID(), diag)
 		if !ok {
 			continue
 		}
@@ -315,13 +316,16 @@ func findSourceError(ctx context.Context, snapshot source.Snapshot, pkgID string
 func diagnosticToAnalyzer(snapshot source.Snapshot, src, msg string) (analyzer *source.Analyzer) {
 	// Make sure that the analyzer we found is enabled.
 	defer func() {
-		if analyzer != nil && !analyzer.Enabled(snapshot) {
+		if analyzer != nil && !analyzer.IsEnabled(snapshot.View()) {
 			analyzer = nil
 		}
 	}()
 	if a, ok := snapshot.View().Options().DefaultAnalyzers[src]; ok {
 		return &a
 	}
+	if a, ok := snapshot.View().Options().StaticcheckAnalyzers[src]; ok {
+		return &a
+	}
 	if a, ok := snapshot.View().Options().ConvenienceAnalyzers[src]; ok {
 		return &a
 	}
@@ -341,10 +345,10 @@ func diagnosticToAnalyzer(snapshot source.Snapshot, src, msg string) (analyzer *
 	return nil
 }
 
-func convenienceFixes(ctx context.Context, snapshot source.Snapshot, ph source.PackageHandle, uri span.URI, rng protocol.Range) ([]protocol.CodeAction, error) {
+func convenienceFixes(ctx context.Context, snapshot source.Snapshot, pkg source.Package, uri span.URI, rng protocol.Range) ([]protocol.CodeAction, error) {
 	var analyzers []*analysis.Analyzer
 	for _, a := range snapshot.View().Options().ConvenienceAnalyzers {
-		if !a.Enabled(snapshot) {
+		if !a.IsEnabled(snapshot.View()) {
 			continue
 		}
 		if a.Command == nil {
@@ -353,7 +357,7 @@ func convenienceFixes(ctx context.Context, snapshot source.Snapshot, ph source.P
 		}
 		analyzers = append(analyzers, a.Analyzer)
 	}
-	diagnostics, err := snapshot.Analyze(ctx, ph.ID(), analyzers...)
+	diagnostics, err := snapshot.Analyze(ctx, pkg.ID(), analyzers...)
 	if err != nil {
 		return nil, err
 	}
@@ -364,7 +368,8 @@ func convenienceFixes(ctx context.Context, snapshot source.Snapshot, ph source.P
 		if d.URI != uri {
 			continue
 		}
-		if d.Range.Start.Line != rng.Start.Line {
+
+		if !protocol.Intersect(d.Range, rng) {
 			continue
 		}
 		action, err := diagnosticToCommandCodeAction(ctx, snapshot, d, nil, protocol.RefactorRewrite)
@@ -406,7 +411,7 @@ func diagnosticToCommandCodeAction(ctx context.Context, snapshot source.Snapshot
 	}, nil
 }
 
-func extractionFixes(ctx context.Context, snapshot source.Snapshot, ph source.PackageHandle, uri span.URI, rng protocol.Range) ([]protocol.CodeAction, error) {
+func extractionFixes(ctx context.Context, snapshot source.Snapshot, pkg source.Package, uri span.URI, rng protocol.Range) ([]protocol.CodeAction, error) {
 	if rng.Start == rng.End {
 		return nil, nil
 	}
@@ -430,7 +435,7 @@ func extractionFixes(ctx context.Context, snapshot source.Snapshot, ph source.Pa
 			Title: command.Title,
 			Kind:  protocol.RefactorExtract,
 			Command: &protocol.Command{
-				Command:   source.CommandExtractFunction.Name,
+				Command:   command.Name,
 				Arguments: jsonArgs,
 			},
 		})
@@ -438,7 +443,7 @@ func extractionFixes(ctx context.Context, snapshot source.Snapshot, ph source.Pa
 	return actions, nil
 }
 
-func documentChanges(fh source.FileHandle, edits []protocol.TextEdit) []protocol.TextDocumentEdit {
+func documentChanges(fh source.VersionedFileHandle, edits []protocol.TextEdit) []protocol.TextDocumentEdit {
 	return []protocol.TextDocumentEdit{
 		{
 			TextDocument: protocol.VersionedTextDocumentIdentifier{
@@ -453,20 +458,19 @@ func documentChanges(fh source.FileHandle, edits []protocol.TextEdit) []protocol
 }
 
 func moduleQuickFixes(ctx context.Context, snapshot source.Snapshot, diagnostics []protocol.Diagnostic) ([]protocol.CodeAction, error) {
-	mth, err := snapshot.ModTidyHandle(ctx)
+	modFH, err := snapshot.GetFile(ctx, snapshot.View().ModFile())
+	if err != nil {
+		return nil, err
+	}
+	tidied, err := snapshot.ModTidy(ctx, modFH)
 	if err == source.ErrTmpModfileUnsupported {
 		return nil, nil
 	}
 	if err != nil {
 		return nil, err
 	}
-	errors, err := mth.Tidy(ctx)
-	if err != nil {
-		return nil, err
-	}
-	pmh := mth.ParseModHandle()
 	var quickFixes []protocol.CodeAction
-	for _, e := range errors {
+	for _, e := range tidied.Errors {
 		var diag *protocol.Diagnostic
 		for _, d := range diagnostics {
 			if sameDiagnostic(d, e) {
@@ -485,14 +489,14 @@ func moduleQuickFixes(ctx context.Context, snapshot source.Snapshot, diagnostics
 				Edit:        protocol.WorkspaceEdit{},
 			}
 			for uri, edits := range fix.Edits {
-				if uri != pmh.Mod().URI() {
+				if uri != modFH.URI() {
 					continue
 				}
 				action.Edit.DocumentChanges = append(action.Edit.DocumentChanges, protocol.TextDocumentEdit{
 					TextDocument: protocol.VersionedTextDocumentIdentifier{
-						Version: pmh.Mod().Version(),
+						Version: modFH.Version(),
 						TextDocumentIdentifier: protocol.TextDocumentIdentifier{
-							URI: protocol.URIFromSpanURI(pmh.Mod().URI()),
+							URI: protocol.URIFromSpanURI(modFH.URI()),
 						},
 					},
 					Edits: edits,
@@ -509,25 +513,21 @@ func sameDiagnostic(d protocol.Diagnostic, e source.Error) bool {
 }
 
 func goModTidy(ctx context.Context, snapshot source.Snapshot) (*protocol.CodeAction, error) {
-	mth, err := snapshot.ModTidyHandle(ctx)
+	modFH, err := snapshot.GetFile(ctx, snapshot.View().ModFile())
 	if err != nil {
 		return nil, err
 	}
-	uri := mth.ParseModHandle().Mod().URI()
-	_, m, _, err := mth.ParseModHandle().Parse(ctx)
+	tidied, err := snapshot.ModTidy(ctx, modFH)
 	if err != nil {
 		return nil, err
 	}
-	left, err := mth.ParseModHandle().Mod().Read()
+	left, err := modFH.Read()
 	if err != nil {
 		return nil, err
 	}
-	right, err := mth.TidiedContent(ctx)
-	if err != nil {
-		return nil, err
-	}
-	edits := snapshot.View().Options().ComputeEdits(uri, string(left), string(right))
-	protocolEdits, err := source.ToProtocolEdits(m, edits)
+	right := tidied.TidiedContent
+	edits := snapshot.View().Options().ComputeEdits(modFH.URI(), string(left), string(right))
+	protocolEdits, err := source.ToProtocolEdits(tidied.Parsed.Mapper, edits)
 	if err != nil {
 		return nil, err
 	}
@@ -537,9 +537,9 @@ func goModTidy(ctx context.Context, snapshot source.Snapshot) (*protocol.CodeAct
 		Edit: protocol.WorkspaceEdit{
 			DocumentChanges: []protocol.TextDocumentEdit{{
 				TextDocument: protocol.VersionedTextDocumentIdentifier{
-					Version: mth.ParseModHandle().Mod().Version(),
+					Version: modFH.Version(),
 					TextDocumentIdentifier: protocol.TextDocumentIdentifier{
-						URI: protocol.URIFromSpanURI(uri),
+						URI: protocol.URIFromSpanURI(modFH.URI()),
 					},
 				},
 				Edits: protocolEdits,
