@@ -11,7 +11,6 @@ import (
 	"regexp"
 	"time"
 
-	"golang.org/x/tools/go/analysis"
 	"golang.org/x/tools/go/analysis/passes/asmdecl"
 	"golang.org/x/tools/go/analysis/passes/assign"
 	"golang.org/x/tools/go/analysis/passes/atomic"
@@ -47,7 +46,6 @@ import (
 	"golang.org/x/tools/internal/lsp/analysis/simplifyslice"
 	"golang.org/x/tools/internal/lsp/analysis/undeclaredname"
 	"golang.org/x/tools/internal/lsp/analysis/unusedparams"
-	"golang.org/x/tools/internal/lsp/debug/tag"
 	"golang.org/x/tools/internal/lsp/diff"
 	"golang.org/x/tools/internal/lsp/diff/myers"
 	"golang.org/x/tools/internal/lsp/protocol"
@@ -100,9 +98,13 @@ func DefaultOptions() Options {
 			CompletionDocumentation: true,
 			EnabledCodeLens: map[string]bool{
 				CommandGenerate.Name:          true,
-				CommandUpgradeDependency.Name: true,
 				CommandRegenerateCgo.Name:     true,
+				CommandTidy.Name:              true,
+				CommandToggleDetails.Name:     false,
+				CommandUpgradeDependency.Name: true,
+				CommandVendor.Name:            true,
 			},
+			ExpandWorkspaceToModule: true,
 		},
 		DebuggingOptions: DebuggingOptions{
 			CompletionBudget:   100 * time.Millisecond,
@@ -117,6 +119,7 @@ func DefaultOptions() Options {
 			DefaultAnalyzers:     defaultAnalyzers(),
 			TypeErrorAnalyzers:   typeErrorAnalyzers(),
 			ConvenienceAnalyzers: convenienceAnalyzers(),
+			StaticcheckAnalyzers: map[string]Analyzer{},
 			GoDiff:               true,
 		},
 	}
@@ -225,6 +228,10 @@ type UserOptions struct {
 
 	// Gofumpt indicates if we should run gofumpt formatting.
 	Gofumpt bool
+
+	// ExpandWorkspaceToModule is true if we should expand the scope of the
+	// workspace to include the modules containing the workspace folders.
+	ExpandWorkspaceToModule bool
 }
 
 type ImportShortcut int
@@ -243,17 +250,6 @@ func (s ImportShortcut) ShowDefinition() bool {
 	return s == Both || s == Definition
 }
 
-type completionOptions struct {
-	deepCompletion    bool
-	unimported        bool
-	documentation     bool
-	fullDocumentation bool
-	placeholders      bool
-	literal           bool
-	matcher           Matcher
-	budget            time.Duration
-}
-
 // Hooks contains configuration that is provided to the Gopls command by the
 // main package.
 type Hooks struct {
@@ -263,13 +259,10 @@ type Hooks struct {
 	DefaultAnalyzers     map[string]Analyzer
 	TypeErrorAnalyzers   map[string]Analyzer
 	ConvenienceAnalyzers map[string]Analyzer
+	StaticcheckAnalyzers map[string]Analyzer
 	GofumptFormat        func(ctx context.Context, src []byte) ([]byte, error)
 }
 
-func (o Options) AddDefaultAnalyzer(a *analysis.Analyzer) {
-	o.DefaultAnalyzers[a.Name] = Analyzer{Analyzer: a, enabled: true}
-}
-
 // ExperimentalOptions defines configuration for features under active
 // development. WARNING: This configuration will be changed in the future. It
 // only exists while these features are under development.
@@ -280,6 +273,14 @@ type ExperimentalOptions struct {
 	// VerboseWorkDoneProgress controls whether the LSP server should send
 	// progress reports for all work done outside the scope of an RPC.
 	VerboseWorkDoneProgress bool
+
+	// Annotations suppress various kinds of optimization diagnostics
+	// that would be reported by the gc_details command.
+	//   noNilcheck suppresses display of nilchecks.
+	//   noEscape suppresses escape choices.
+	//   noInline suppresses inlining choices.
+	//   noBounds suppresses bounds checking diagnositcs.
+	Annotations map[string]bool
 }
 
 // DebuggingOptions should not affect the logical execution of Gopls, but may
@@ -480,8 +481,10 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 			o.SymbolStyle = FullyQualifiedSymbols
 		case "dynamic":
 			o.SymbolStyle = DynamicSymbols
-		default:
+		case "package":
 			o.SymbolStyle = PackageQualifiedSymbols
+		default:
+			result.errorf("Unsupported symbol style %q", style)
 		}
 
 	case "hoverKind":
@@ -501,7 +504,7 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 		case "Structured":
 			o.HoverKind = Structured
 		default:
-			result.errorf("Unsupported hover kind", tag.HoverKind.Of(hoverKind))
+			result.errorf("Unsupported hover kind %q", hoverKind)
 		}
 
 	case "linkTarget":
@@ -525,6 +528,18 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 	case "analyses":
 		result.setBoolMap(&o.UserEnabledAnalyses)
 
+	case "annotations":
+		result.setBoolMap(&o.Annotations)
+		for k := range o.Annotations {
+			switch k {
+			case "noEscape", "noNilcheck", "noInline", "noBounds":
+				continue
+			default:
+				result.Name += ":" + k // put mistake(s) in the message
+				result.State = OptionUnexpected
+			}
+		}
+
 	case "codelens":
 		var lensOverrides map[string]bool
 		result.setBoolMap(&lensOverrides)
@@ -555,6 +570,9 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 	case "gofumpt":
 		result.setBool(&o.Gofumpt)
 
+	case "expandWorkspaceToModule":
+		result.setBool(&o.ExpandWorkspaceToModule)
+
 	// Replaced settings.
 	case "experimentalDisabledAnalyses":
 		result.State = OptionDeprecated
@@ -659,17 +677,22 @@ func (r *OptionResult) setString(s *string) {
 // snapshot.
 func EnabledAnalyzers(snapshot Snapshot) (analyzers []Analyzer) {
 	for _, a := range snapshot.View().Options().DefaultAnalyzers {
-		if a.Enabled(snapshot) {
+		if a.IsEnabled(snapshot.View()) {
 			analyzers = append(analyzers, a)
 		}
 	}
 	for _, a := range snapshot.View().Options().TypeErrorAnalyzers {
-		if a.Enabled(snapshot) {
+		if a.IsEnabled(snapshot.View()) {
 			analyzers = append(analyzers, a)
 		}
 	}
 	for _, a := range snapshot.View().Options().ConvenienceAnalyzers {
-		if a.Enabled(snapshot) {
+		if a.IsEnabled(snapshot.View()) {
+			analyzers = append(analyzers, a)
+		}
+	}
+	for _, a := range snapshot.View().Options().StaticcheckAnalyzers {
+		if a.IsEnabled(snapshot.View()) {
 			analyzers = append(analyzers, a)
 		}
 	}
@@ -682,22 +705,23 @@ func typeErrorAnalyzers() map[string]Analyzer {
 			Analyzer:       fillreturns.Analyzer,
 			FixesError:     fillreturns.FixesError,
 			HighConfidence: true,
+			Enabled:        true,
 		},
 		nonewvars.Analyzer.Name: {
 			Analyzer:   nonewvars.Analyzer,
 			FixesError: nonewvars.FixesError,
-			enabled:    true,
+			Enabled:    true,
 		},
 		noresultvalues.Analyzer.Name: {
 			Analyzer:   noresultvalues.Analyzer,
 			FixesError: noresultvalues.FixesError,
-			enabled:    true,
+			Enabled:    true,
 		},
 		undeclaredname.Analyzer.Name: {
 			Analyzer:   undeclaredname.Analyzer,
 			FixesError: undeclaredname.FixesError,
 			Command:    CommandUndeclaredName,
-			enabled:    true,
+			Enabled:    true,
 		},
 	}
 }
@@ -707,7 +731,7 @@ func convenienceAnalyzers() map[string]Analyzer {
 		fillstruct.Analyzer.Name: {
 			Analyzer: fillstruct.Analyzer,
 			Command:  CommandFillStruct,
-			enabled:  true,
+			Enabled:  true,
 		},
 	}
 }
@@ -715,40 +739,40 @@ func convenienceAnalyzers() map[string]Analyzer {
 func defaultAnalyzers() map[string]Analyzer {
 	return map[string]Analyzer{
 		// The traditional vet suite:
-		asmdecl.Analyzer.Name:      {Analyzer: asmdecl.Analyzer, enabled: true},
-		assign.Analyzer.Name:       {Analyzer: assign.Analyzer, enabled: true},
-		atomic.Analyzer.Name:       {Analyzer: atomic.Analyzer, enabled: true},
-		atomicalign.Analyzer.Name:  {Analyzer: atomicalign.Analyzer, enabled: true},
-		bools.Analyzer.Name:        {Analyzer: bools.Analyzer, enabled: true},
-		buildtag.Analyzer.Name:     {Analyzer: buildtag.Analyzer, enabled: true},
-		cgocall.Analyzer.Name:      {Analyzer: cgocall.Analyzer, enabled: true},
-		composite.Analyzer.Name:    {Analyzer: composite.Analyzer, enabled: true},
-		copylock.Analyzer.Name:     {Analyzer: copylock.Analyzer, enabled: true},
-		errorsas.Analyzer.Name:     {Analyzer: errorsas.Analyzer, enabled: true},
-		httpresponse.Analyzer.Name: {Analyzer: httpresponse.Analyzer, enabled: true},
-		loopclosure.Analyzer.Name:  {Analyzer: loopclosure.Analyzer, enabled: true},
-		lostcancel.Analyzer.Name:   {Analyzer: lostcancel.Analyzer, enabled: true},
-		nilfunc.Analyzer.Name:      {Analyzer: nilfunc.Analyzer, enabled: true},
-		printf.Analyzer.Name:       {Analyzer: printf.Analyzer, enabled: true},
-		shift.Analyzer.Name:        {Analyzer: shift.Analyzer, enabled: true},
-		stdmethods.Analyzer.Name:   {Analyzer: stdmethods.Analyzer, enabled: true},
-		structtag.Analyzer.Name:    {Analyzer: structtag.Analyzer, enabled: true},
-		tests.Analyzer.Name:        {Analyzer: tests.Analyzer, enabled: true},
-		unmarshal.Analyzer.Name:    {Analyzer: unmarshal.Analyzer, enabled: true},
-		unreachable.Analyzer.Name:  {Analyzer: unreachable.Analyzer, enabled: true},
-		unsafeptr.Analyzer.Name:    {Analyzer: unsafeptr.Analyzer, enabled: true},
-		unusedresult.Analyzer.Name: {Analyzer: unusedresult.Analyzer, enabled: true},
+		asmdecl.Analyzer.Name:      {Analyzer: asmdecl.Analyzer, Enabled: true},
+		assign.Analyzer.Name:       {Analyzer: assign.Analyzer, Enabled: true},
+		atomic.Analyzer.Name:       {Analyzer: atomic.Analyzer, Enabled: true},
+		atomicalign.Analyzer.Name:  {Analyzer: atomicalign.Analyzer, Enabled: true},
+		bools.Analyzer.Name:        {Analyzer: bools.Analyzer, Enabled: true},
+		buildtag.Analyzer.Name:     {Analyzer: buildtag.Analyzer, Enabled: true},
+		cgocall.Analyzer.Name:      {Analyzer: cgocall.Analyzer, Enabled: true},
+		composite.Analyzer.Name:    {Analyzer: composite.Analyzer, Enabled: true},
+		copylock.Analyzer.Name:     {Analyzer: copylock.Analyzer, Enabled: true},
+		errorsas.Analyzer.Name:     {Analyzer: errorsas.Analyzer, Enabled: true},
+		httpresponse.Analyzer.Name: {Analyzer: httpresponse.Analyzer, Enabled: true},
+		loopclosure.Analyzer.Name:  {Analyzer: loopclosure.Analyzer, Enabled: true},
+		lostcancel.Analyzer.Name:   {Analyzer: lostcancel.Analyzer, Enabled: true},
+		nilfunc.Analyzer.Name:      {Analyzer: nilfunc.Analyzer, Enabled: true},
+		printf.Analyzer.Name:       {Analyzer: printf.Analyzer, Enabled: true},
+		shift.Analyzer.Name:        {Analyzer: shift.Analyzer, Enabled: true},
+		stdmethods.Analyzer.Name:   {Analyzer: stdmethods.Analyzer, Enabled: true},
+		structtag.Analyzer.Name:    {Analyzer: structtag.Analyzer, Enabled: true},
+		tests.Analyzer.Name:        {Analyzer: tests.Analyzer, Enabled: true},
+		unmarshal.Analyzer.Name:    {Analyzer: unmarshal.Analyzer, Enabled: true},
+		unreachable.Analyzer.Name:  {Analyzer: unreachable.Analyzer, Enabled: true},
+		unsafeptr.Analyzer.Name:    {Analyzer: unsafeptr.Analyzer, Enabled: true},
+		unusedresult.Analyzer.Name: {Analyzer: unusedresult.Analyzer, Enabled: true},
 
 		// Non-vet analyzers:
-		deepequalerrors.Analyzer.Name:  {Analyzer: deepequalerrors.Analyzer, enabled: true},
-		sortslice.Analyzer.Name:        {Analyzer: sortslice.Analyzer, enabled: true},
-		testinggoroutine.Analyzer.Name: {Analyzer: testinggoroutine.Analyzer, enabled: true},
-		unusedparams.Analyzer.Name:     {Analyzer: unusedparams.Analyzer, enabled: false},
+		deepequalerrors.Analyzer.Name:  {Analyzer: deepequalerrors.Analyzer, Enabled: true},
+		sortslice.Analyzer.Name:        {Analyzer: sortslice.Analyzer, Enabled: true},
+		testinggoroutine.Analyzer.Name: {Analyzer: testinggoroutine.Analyzer, Enabled: true},
+		unusedparams.Analyzer.Name:     {Analyzer: unusedparams.Analyzer, Enabled: false},
 
 		// gofmt -s suite:
-		simplifycompositelit.Analyzer.Name: {Analyzer: simplifycompositelit.Analyzer, enabled: true, HighConfidence: true},
-		simplifyrange.Analyzer.Name:        {Analyzer: simplifyrange.Analyzer, enabled: true, HighConfidence: true},
-		simplifyslice.Analyzer.Name:        {Analyzer: simplifyslice.Analyzer, enabled: true, HighConfidence: true},
+		simplifycompositelit.Analyzer.Name: {Analyzer: simplifycompositelit.Analyzer, Enabled: true, HighConfidence: true},
+		simplifyrange.Analyzer.Name:        {Analyzer: simplifyrange.Analyzer, Enabled: true, HighConfidence: true},
+		simplifyslice.Analyzer.Name:        {Analyzer: simplifyslice.Analyzer, Enabled: true, HighConfidence: true},
 	}
 }
 
