@@ -29,13 +29,16 @@ type Snapshot interface {
 	// View returns the View associated with this snapshot.
 	View() View
 
+	// Fileset returns the Fileset used to parse all the Go files in this snapshot.
+	FileSet() *token.FileSet
+
 	// FindFile returns the FileHandle for the given URI, if it is already
 	// in the given snapshot.
-	FindFile(uri span.URI) FileHandle
+	FindFile(uri span.URI) VersionedFileHandle
 
 	// GetFile returns the FileHandle for a given URI, initializing it
 	// if it is not already part of the snapshot.
-	GetFile(ctx context.Context, uri span.URI) (FileHandle, error)
+	GetFile(ctx context.Context, uri span.URI) (VersionedFileHandle, error)
 
 	// IsOpen returns whether the editor currently has a file open.
 	IsOpen(uri span.URI) bool
@@ -43,6 +46,21 @@ type Snapshot interface {
 	// IsSaved returns whether the contents are saved on disk or not.
 	IsSaved(uri span.URI) bool
 
+	// ParseGo returns the parsed AST for the file.
+	// If the file is not available, returns nil and an error.
+	ParseGo(ctx context.Context, fh FileHandle, mode ParseMode) (*ParsedGoFile, error)
+
+	// PosToField is a cache of *ast.Fields by token.Pos. This allows us
+	// to quickly find corresponding *ast.Field node given a *types.Var.
+	// We must refer to the AST to render type aliases properly when
+	// formatting signatures and other types.
+	PosToField(ctx context.Context, pgf *ParsedGoFile) (map[token.Pos]*ast.Field, error)
+
+	// PosToDecl maps certain objects' positions to their surrounding
+	// ast.Decl. This mapping is used when building the documentation
+	// string for the objects.
+	PosToDecl(ctx context.Context, pgf *ParsedGoFile) (map[token.Pos]ast.Decl, error)
+
 	// Analyze runs the analyses for the given package at this snapshot.
 	Analyze(ctx context.Context, pkgID string, analyzers ...*analysis.Analyzer) ([]*Error, error)
 
@@ -58,63 +76,47 @@ type Snapshot interface {
 	// -modfile flag.
 	RunGoCommandDirect(ctx context.Context, verb string, args []string) error
 
-	// ParseModHandle is used to parse go.mod files.
-	ParseModHandle(ctx context.Context, fh FileHandle) (ParseModHandle, error)
+	// ParseMod is used to parse go.mod files.
+	ParseMod(ctx context.Context, fh FileHandle) (*ParsedModule, error)
 
-	// ModWhyHandle is used get the results of `go mod why` for a given module.
-	// It only works for go.mod files that can be parsed, hence it takes a
-	// ParseModHandle.
-	ModWhyHandle(ctx context.Context) (ModWhyHandle, error)
+	// ModWhy returns the results of `go mod why` for the module specified by
+	// the given go.mod file.
+	ModWhy(ctx context.Context, fh FileHandle) (map[string]string, error)
 
-	// ModWhyHandle is used get the possible upgrades for the dependencies of
-	// a given module. It only works for go.mod files that can be parsed, hence
-	// it takes a ParseModHandle.
-	ModUpgradeHandle(ctx context.Context) (ModUpgradeHandle, error)
+	// ModUpgrade returns the possible updates for the module specified by the
+	// given go.mod file.
+	ModUpgrade(ctx context.Context, fh FileHandle) (map[string]string, error)
 
-	// ModWhyHandle is used get the results of `go mod tidy` for a given
-	// module. It only works for go.mod files that can be parsed, hence it
-	// takes a ParseModHandle.
-	ModTidyHandle(ctx context.Context) (ModTidyHandle, error)
+	// ModTidy returns the results of `go mod tidy` for the module specified by
+	// the given go.mod file.
+	ModTidy(ctx context.Context, fh FileHandle) (*TidiedModule, error)
 
-	// PackageHandles returns the PackageHandles for the packages that this file
-	// belongs to.
-	PackageHandles(ctx context.Context, fh FileHandle) ([]PackageHandle, error)
+	// BuiltinPackage returns information about the special builtin package.
+	BuiltinPackage(ctx context.Context) (*BuiltinPackage, error)
+
+	// PackagesForFile returns the packages that this file belongs to, checked
+	// in mode.
+	PackagesForFile(ctx context.Context, uri span.URI, mode TypecheckMode) ([]Package, error)
 
 	// GetActiveReverseDeps returns the active files belonging to the reverse
-	// dependencies of this file's package.
-	GetReverseDependencies(ctx context.Context, id string) ([]PackageHandle, error)
+	// dependencies of this file's package, checked in TypecheckWorkspace mode.
+	GetReverseDependencies(ctx context.Context, id string) ([]Package, error)
 
-	// CachedImportPaths returns all the imported packages loaded in this snapshot,
-	// indexed by their import path.
+	// CachedImportPaths returns all the imported packages loaded in this
+	// snapshot, indexed by their import path and checked in TypecheckWorkspace
+	// mode.
 	CachedImportPaths(ctx context.Context) (map[string]Package, error)
 
-	// KnownPackages returns all the packages loaded in this snapshot.
-	// Workspace packages may be parsed in ParseFull mode, whereas transitive
-	// dependencies will be in ParseExported mode.
-	KnownPackages(ctx context.Context) ([]PackageHandle, error)
+	// KnownPackages returns all the packages loaded in this snapshot, checked
+	// in TypecheckWorkspace mode.
+	KnownPackages(ctx context.Context) ([]Package, error)
 
-	// WorkspacePackages returns the PackageHandles for the snapshot's
-	// top-level packages.
-	WorkspacePackages(ctx context.Context) ([]PackageHandle, error)
-}
+	// WorkspacePackages returns the snapshot's top-level packages.
+	WorkspacePackages(ctx context.Context) ([]Package, error)
 
-// PackageHandle represents a handle to a specific version of a package.
-// It is uniquely defined by the file handles that make up the package.
-type PackageHandle interface {
-	// ID returns the ID of the package associated with the PackageHandle.
-	ID() string
-
-	// CompiledGoFiles returns the ParseGoHandles composing the package.
-	CompiledGoFiles() []ParseGoHandle
-
-	// Check returns the type-checked Package for the PackageHandle.
-	Check(ctx context.Context) (Package, error)
-
-	// Cached returns the Package for the PackageHandle if it has already been stored.
-	Cached() (Package, error)
-
-	// MissingDependencies reports any unresolved imports.
-	MissingDependencies() []string
+	// WorkspaceDirectories returns any directory known by the view. For views
+	// within a module, this is the module root and any replace targets.
+	WorkspaceDirectories(ctx context.Context) []span.URI
 }
 
 // View represents a single workspace.
@@ -133,16 +135,16 @@ type View interface {
 	// ModFile is the go.mod file at the root of this view. It may not exist.
 	ModFile() span.URI
 
-	// BuiltinPackage returns the go/ast.Object for the given name in the builtin package.
-	BuiltinPackage(ctx context.Context) (BuiltinPackage, error)
-
 	// BackgroundContext returns a context used for all background processing
 	// on behalf of this view.
 	BackgroundContext() context.Context
 
-	// Shutdown closes this view, and detaches it from it's session.
+	// Shutdown closes this view, and detaches it from its session.
 	Shutdown(ctx context.Context)
 
+	// AwaitInitialized waits until a view is initialized
+	AwaitInitialized(ctx context.Context)
+
 	// WriteEnv writes the view-specific environment to the io.Writer.
 	WriteEnv(ctx context.Context, w io.Writer) error
 
@@ -160,10 +162,10 @@ type View interface {
 	SetOptions(context.Context, Options) (View, error)
 
 	// Snapshot returns the current snapshot for the view.
-	Snapshot() Snapshot
+	Snapshot(ctx context.Context) (Snapshot, func())
 
 	// Rebuild rebuilds the current view, replacing the original view in its session.
-	Rebuild(ctx context.Context) (Snapshot, error)
+	Rebuild(ctx context.Context) (Snapshot, func(), error)
 
 	// InvalidBuildConfiguration returns true if there is some error in the
 	// user's workspace. In particular, if they are both outside of a module
@@ -177,15 +179,41 @@ type View interface {
 	// IgnoredFile reports if a file would be ignored by a `go list` of the whole
 	// workspace.
 	IgnoredFile(uri span.URI) bool
-
-	// WorkspaceDirectories returns any directory known by the view. For views
-	// within a module, this is the module root and any replace targets.
-	WorkspaceDirectories(ctx context.Context) ([]string, error)
 }
 
-type BuiltinPackage interface {
-	Package() *ast.Package
-	ParseGoHandle() ParseGoHandle
+type BuiltinPackage struct {
+	Package    *ast.Package
+	ParsedFile *ParsedGoFile
+}
+
+// A ParsedGoFile contains the results of parsing a Go file.
+type ParsedGoFile struct {
+	URI  span.URI
+	Mode ParseMode
+	File *ast.File
+	Tok  *token.File
+	// Source code used to build the AST. It may be different from the
+	// actual content of the file if we have fixed the AST.
+	Src      []byte
+	Mapper   *protocol.ColumnMapper
+	ParseErr error
+}
+
+// A ParsedModule contains the results of parsing a go.mod file.
+type ParsedModule struct {
+	File        *modfile.File
+	Mapper      *protocol.ColumnMapper
+	ParseErrors []Error
+}
+
+// A TidiedModule contains the results of running `go mod tidy` on a module.
+type TidiedModule struct {
+	// The parsed module, which is guaranteed to have parsed successfully.
+	Parsed *ParsedModule
+	// Diagnostics representing changes made by `go mod tidy`.
+	Errors []Error
+	// The bytes of the go.mod file after it was tidied.
+	TidiedContent []byte
 }
 
 // Session represents a single connection from a client.
@@ -193,11 +221,11 @@ type BuiltinPackage interface {
 // of the client.
 // A session may have many active views at any given time.
 type Session interface {
-	// NewView creates a new View and returns it.
-	NewView(ctx context.Context, name string, folder span.URI, options Options) (View, Snapshot, error)
+	// NewView creates a new View, returning it and its first snapshot.
+	NewView(ctx context.Context, name string, folder span.URI, options Options) (View, Snapshot, func(), error)
 
-	// Cache returns the cache that created this session.
-	Cache() Cache
+	// Cache returns the cache that created this session, for debugging only.
+	Cache() interface{}
 
 	// View returns a view with a matching name, if the session has one.
 	View(name string) View
@@ -214,9 +242,9 @@ type Session interface {
 	// GetFile returns a handle for the specified file.
 	GetFile(ctx context.Context, uri span.URI) (FileHandle, error)
 
-	// DidModifyFile reports a file modification to the session.
-	// It returns the resulting snapshots, a guaranteed one per view.
-	DidModifyFiles(ctx context.Context, changes []FileModification) ([]Snapshot, error)
+	// DidModifyFile reports a file modification to the session. It returns the
+	// resulting snapshots, a guaranteed one per view.
+	DidModifyFiles(ctx context.Context, changes []FileModification) ([]Snapshot, []func(), []span.URI, error)
 
 	// Overlays returns a slice of file overlays for the session.
 	Overlays() []Overlay
@@ -230,17 +258,10 @@ type Session interface {
 
 // Overlay is the type for a file held in memory on a session.
 type Overlay interface {
-	// Session returns the session this overlay belongs to.
-	Session() Session
-
-	// Identity returns the FileIdentity for the overlay.
-	Identity() FileIdentity
+	VersionedFileHandle
 
 	// Saved returns whether this overlay has been saved to disk.
 	Saved() bool
-
-	// Data is the contents of the overlay held in memory.
-	Data() []byte
 }
 
 // FileModification represents a modification to a file.
@@ -295,86 +316,6 @@ func (a FileAction) String() string {
 	}
 }
 
-// Cache abstracts the core logic of dealing with the environment from the
-// higher level logic that processes the information to produce results.
-// The cache provides access to files and their contents, so the source
-// package does not directly access the file system.
-// A single cache is intended to be process wide, and is the primary point of
-// sharing between all consumers.
-// A cache may have many active sessions at any given time.
-type Cache interface {
-	// FileSet returns the shared fileset used by all files in the system.
-	FileSet() *token.FileSet
-
-	// GetFile returns a file handle for the given URI.
-	GetFile(ctx context.Context, uri span.URI) (FileHandle, error)
-
-	// ParseGoHandle returns a ParseGoHandle for the given file handle.
-	ParseGoHandle(ctx context.Context, fh FileHandle, mode ParseMode) ParseGoHandle
-}
-
-// ParseGoHandle represents a handle to the AST for a file.
-type ParseGoHandle interface {
-	// File returns a file handle for which to get the AST.
-	File() FileHandle
-
-	// Mode returns the parse mode of this handle.
-	Mode() ParseMode
-
-	// Parse returns the parsed AST for the file.
-	// If the file is not available, returns nil and an error.
-	Parse(ctx context.Context) (file *ast.File, src []byte, m *protocol.ColumnMapper, parseErr error, err error)
-
-	// Cached returns the AST for this handle, if it has already been stored.
-	Cached() (file *ast.File, src []byte, m *protocol.ColumnMapper, parseErr error, err error)
-
-	// PosToField is a cache of *ast.Fields by token.Pos. This allows us
-	// to quickly find corresponding *ast.Field node given a *types.Var.
-	// We must refer to the AST to render type aliases properly when
-	// formatting signatures and other types.
-	PosToField(context.Context) (map[token.Pos]*ast.Field, error)
-
-	// PosToDecl maps certain objects' positions to their surrounding
-	// ast.Decl. This mapping is used when building the documentation
-	// string for the objects.
-	PosToDecl(context.Context) (map[token.Pos]ast.Decl, error)
-}
-
-type ParseModHandle interface {
-	// Mod returns the file handle for the go.mod file.
-	Mod() FileHandle
-
-	// Sum returns the file handle for the analogous go.sum file. It may be nil.
-	Sum() FileHandle
-
-	// Parse returns the parsed go.mod file, a column mapper, and a list of
-	// parse for the go.mod file.
-	Parse(ctx context.Context) (*modfile.File, *protocol.ColumnMapper, []Error, error)
-}
-
-type ModUpgradeHandle interface {
-	// Upgrades returns the latest versions for each of the module's
-	// dependencies.
-	Upgrades(ctx context.Context) (map[string]string, error)
-}
-
-type ModWhyHandle interface {
-	// Why returns the results of `go mod why` for every dependency of the
-	// module.
-	Why(ctx context.Context) (map[string]string, error)
-}
-
-type ModTidyHandle interface {
-	// Mod is the ParseModHandle associated with the go.mod file being tidied.
-	ParseModHandle() ParseModHandle
-
-	// Tidy returns the results of `go mod tidy` for the module.
-	Tidy(ctx context.Context) ([]Error, error)
-
-	// TidiedContent is the content of the tidied go.mod file.
-	TidiedContent(ctx context.Context) ([]byte, error)
-}
-
 var ErrTmpModfileUnsupported = errors.New("-modfile is unsupported for this Go version")
 
 // ParseMode controls the content of the AST produced when parsing a source file.
@@ -383,7 +324,7 @@ type ParseMode int
 const (
 	// ParseHeader specifies that the main package declaration and imports are needed.
 	// This is the mode used when attempting to examine the package graph structure.
-	ParseHeader = ParseMode(iota)
+	ParseHeader ParseMode = iota
 
 	// ParseExported specifies that the public symbols are needed, but things like
 	// private symbols and function bodies are not.
@@ -397,17 +338,52 @@ const (
 	ParseFull
 )
 
+// TypecheckMode controls what kind of parsing should be done (see ParseMode)
+// while type checking a package.
+type TypecheckMode int
+
+const (
+	// Invalid default value.
+	TypecheckUnknown TypecheckMode = iota
+	// TypecheckFull means to use ParseFull.
+	TypecheckFull
+	// TypecheckWorkspace means to use ParseFull for workspace packages, and
+	// ParseExported for others.
+	TypecheckWorkspace
+	// TypecheckAll means ParseFull for workspace packages, and both Full and
+	// Exported for others. Only valid for some functions.
+	TypecheckAll
+)
+
+type VersionedFileHandle interface {
+	FileHandle
+	Version() float64
+	Session() string
+
+	// LSPIdentity returns the version identity of a file.
+	VersionedFileIdentity() VersionedFileIdentity
+}
+
+type VersionedFileIdentity struct {
+	URI span.URI
+
+	// SessionID is the ID of the LSP session.
+	SessionID string
+
+	// Version is the version of the file, as specified by the client. It should
+	// only be set in combination with SessionID.
+	Version float64
+}
+
 // FileHandle represents a handle to a specific version of a single file.
 type FileHandle interface {
 	URI() span.URI
 	Kind() FileKind
-	Version() float64
 
 	// Identity returns a FileIdentity for the file, even if there was an error
 	// reading it.
 	// It is a fatal error to call Identity on a file that has not yet been read.
-	Identity() FileIdentity
-
+	FileIdentity() FileIdentity
 	// Read reads the contents of a file.
 	// If the file is not available, returns a nil slice and an error.
 	Read() ([]byte, error)
@@ -417,27 +393,13 @@ type FileHandle interface {
 type FileIdentity struct {
 	URI span.URI
 
-	// SessionID is the ID of the LSP session.
-	SessionID string
-
-	// Version is the version of the file, as specified by the client. It should
-	// only be set in combination with SessionID.
-	Version float64
-
-	// Identifier represents a unique identifier for the file.
-	// It could be a file's modification time or its SHA1 hash if it is not on disk.
-	Identifier string
+	// Identifier represents a unique identifier for the file's content.
+	Hash string
 
 	// Kind is the file's kind.
 	Kind FileKind
 }
 
-func (fileID FileIdentity) String() string {
-	// Version is not part of the FileIdentity string,
-	// as it can remain change even if the file does not.
-	return fmt.Sprintf("%s%s%s", fileID.URI, fileID.Identifier, fileID.Kind)
-}
-
 // FileKind describes the kind of the file in question.
 // It can be one of Go, mod, or sum.
 type FileKind int
@@ -458,7 +420,11 @@ const (
 // that let the user know how to use the analyzer.
 type Analyzer struct {
 	Analyzer *analysis.Analyzer
-	enabled  bool
+
+	// Enabled reports whether the analyzer is enabled. This value can be
+	// configured per-analysis in user settings. For staticcheck analyzers,
+	// the value of the Staticcheck setting overrides this field.
+	Enabled bool
 
 	// Command is the name of the command used to invoke the suggested fixes
 	// for the analyzer. It is non-nil if we expect this analyzer to provide
@@ -476,11 +442,17 @@ type Analyzer struct {
 	FixesError func(msg string) bool
 }
 
-func (a Analyzer) Enabled(snapshot Snapshot) bool {
-	if enabled, ok := snapshot.View().Options().UserEnabledAnalyses[a.Analyzer.Name]; ok {
+func (a Analyzer) IsEnabled(view View) bool {
+	// Staticcheck analyzers can only be enabled when staticcheck is on.
+	if _, ok := view.Options().StaticcheckAnalyzers[a.Analyzer.Name]; ok {
+		if !view.Options().StaticCheck {
+			return false
+		}
+	}
+	if enabled, ok := view.Options().UserEnabledAnalyses[a.Analyzer.Name]; ok {
 		return enabled
 	}
-	return a.enabled
+	return a.Enabled
 }
 
 // Package represents a Go package that has been type-checked. It maintains
@@ -489,8 +461,8 @@ type Package interface {
 	ID() string
 	Name() string
 	PkgPath() string
-	CompiledGoFiles() []ParseGoHandle
-	File(uri span.URI) (ParseGoHandle, error)
+	CompiledGoFiles() []*ParsedGoFile
+	File(uri span.URI) (*ParsedGoFile, error)
 	GetSyntax() []*ast.File
 	GetErrors() []*Error
 	GetTypes() *types.Package
@@ -499,6 +471,7 @@ type Package interface {
 	IsIllTyped() bool
 	ForTest() string
 	GetImport(pkgPath string) (Package, error)
+	MissingDependencies() []string
 	Imports() []Package
 	Module() *packages.Module
 }
@@ -513,6 +486,9 @@ type Error struct {
 	Related        []RelatedInformation
 }
 
+// GoModTidy is the source for a diagnostic computed by running `go mod tidy`.
+const GoModTidy = "go mod tidy"
+
 type ErrorKind int
 
 const (
