@@ -28,6 +28,7 @@ import (
 	"golang.org/x/tools/go/packages/packagestest"
 	"golang.org/x/tools/internal/lsp/protocol"
 	"golang.org/x/tools/internal/lsp/source"
+	"golang.org/x/tools/internal/lsp/source/completion"
 	"golang.org/x/tools/internal/span"
 	"golang.org/x/tools/internal/testenv"
 	"golang.org/x/tools/txtar"
@@ -43,9 +44,10 @@ const (
 
 var UpdateGolden = flag.Bool("golden", false, "Update golden files")
 
+type CallHierarchy map[span.Span]*CallHierarchyResult
 type CodeLens map[span.URI][]protocol.CodeLens
 type Diagnostics map[span.URI][]*source.Diagnostic
-type CompletionItems map[token.Pos]*source.CompletionItem
+type CompletionItems map[token.Pos]*completion.CompletionItem
 type Completions map[span.Span][]Completion
 type CompletionSnippets map[span.Span][]CompletionSnippet
 type UnimportedCompletions map[span.Span][]Completion
@@ -74,6 +76,7 @@ type Links map[span.URI][]Link
 type Data struct {
 	Config                        packages.Config
 	Exported                      *packagestest.Exported
+	CallHierarchy                 CallHierarchy
 	CodeLens                      CodeLens
 	Diagnostics                   Diagnostics
 	CompletionItems               CompletionItems
@@ -117,6 +120,7 @@ type Data struct {
 }
 
 type Tests interface {
+	CallHierarchy(*testing.T, span.Span, *CallHierarchyResult)
 	CodeLens(*testing.T, span.URI, []protocol.CodeLens)
 	Diagnostics(*testing.T, span.URI, []*source.Diagnostic)
 	Completion(*testing.T, span.Span, Completion, CompletionItems)
@@ -197,6 +201,10 @@ type CompletionSnippet struct {
 	PlaceholderSnippet string
 }
 
+type CallHierarchyResult struct {
+	IncomingCalls, OutgoingCalls []protocol.CallHierarchyItem
+}
+
 type Link struct {
 	Src          span.Span
 	Target       string
@@ -274,6 +282,7 @@ func Load(t testing.TB, exporter packagestest.Exporter, dir string) []*Data {
 	var data []*Data
 	for _, folder := range folders {
 		datum := &Data{
+			CallHierarchy:                 make(CallHierarchy),
 			CodeLens:                      make(CodeLens),
 			Diagnostics:                   make(Diagnostics),
 			CompletionItems:               make(CompletionItems),
@@ -425,6 +434,8 @@ func Load(t testing.TB, exporter packagestest.Exporter, dir string) []*Data {
 			"link":            datum.collectLinks,
 			"suggestedfix":    datum.collectSuggestedFixes,
 			"extractfunc":     datum.collectFunctionExtractions,
+			"incomingcalls":   datum.collectIncomingCalls,
+			"outgoingcalls":   datum.collectOutgoingCalls,
 		}); err != nil {
 			t.Fatal(err)
 		}
@@ -495,6 +506,16 @@ func Run(t *testing.T, tests Tests, data *Data) {
 		}
 	}
 
+	t.Run("CallHierarchy", func(t *testing.T) {
+		t.Helper()
+		for spn, callHierarchyResult := range data.CallHierarchy {
+			t.Run(SpanName(spn), func(t *testing.T) {
+				t.Helper()
+				tests.CallHierarchy(t, spn, callHierarchyResult)
+			})
+		}
+	})
+
 	t.Run("Completion", func(t *testing.T) {
 		t.Helper()
 		eachCompletion(t, data.Completions, tests.Completion)
@@ -807,6 +828,7 @@ func checkData(t *testing.T, data *Data) {
 		return count
 	}
 
+	fmt.Fprintf(buf, "CallHierarchyCount = %v\n", len(data.CallHierarchy))
 	fmt.Fprintf(buf, "CodeLensCount = %v\n", countCodeLens(data.CodeLens))
 	fmt.Fprintf(buf, "CompletionsCount = %v\n", countCompletions(data.Completions))
 	fmt.Fprintf(buf, "CompletionSnippetCount = %v\n", snippetCount)
@@ -1016,7 +1038,7 @@ func (data *Data) collectCompletionItems(pos token.Pos, args []string) {
 	if len(args) == 4 {
 		documentation = args[3]
 	}
-	data.CompletionItems[pos] = &source.CompletionItem{
+	data.CompletionItems[pos] = &completion.CompletionItem{
 		Label:         label,
 		Detail:        detail,
 		Kind:          protocol.ParseCompletionItemKind(kind),
@@ -1060,6 +1082,60 @@ func (data *Data) collectImplementations(src span.Span, targets []span.Span) {
 	data.Implementations[src] = targets
 }
 
+func (data *Data) collectIncomingCalls(src span.Span, calls []span.Span) {
+	for _, call := range calls {
+		m, err := data.Mapper(call.URI())
+		if err != nil {
+			data.t.Fatal(err)
+		}
+		rng, err := m.Range(call)
+		if err != nil {
+			data.t.Fatal(err)
+		}
+		// we're only comparing protocol.range
+		if data.CallHierarchy[src] != nil {
+			data.CallHierarchy[src].IncomingCalls = append(data.CallHierarchy[src].IncomingCalls,
+				protocol.CallHierarchyItem{
+					URI:   protocol.DocumentURI(call.URI()),
+					Range: rng,
+				})
+		} else {
+			data.CallHierarchy[src] = &CallHierarchyResult{
+				IncomingCalls: []protocol.CallHierarchyItem{
+					{URI: protocol.DocumentURI(call.URI()), Range: rng},
+				},
+			}
+		}
+	}
+}
+
+func (data *Data) collectOutgoingCalls(src span.Span, calls []span.Span) {
+	for _, call := range calls {
+		m, err := data.Mapper(call.URI())
+		if err != nil {
+			data.t.Fatal(err)
+		}
+		rng, err := m.Range(call)
+		if err != nil {
+			data.t.Fatal(err)
+		}
+		// we're only comparing protocol.range
+		if data.CallHierarchy[src] != nil {
+			data.CallHierarchy[src].OutgoingCalls = append(data.CallHierarchy[src].OutgoingCalls,
+				protocol.CallHierarchyItem{
+					URI:   protocol.DocumentURI(call.URI()),
+					Range: rng,
+				})
+		} else {
+			data.CallHierarchy[src] = &CallHierarchyResult{
+				OutgoingCalls: []protocol.CallHierarchyItem{
+					{URI: protocol.DocumentURI(call.URI()), Range: rng},
+				},
+			}
+		}
+	}
+}
+
 func (data *Data) collectHoverDefinitions(src, target span.Span) {
 	data.Definitions[src] = Definition{
 		Src:       src,
